/******************************************************************************
 *
 * parser.y
 *
 *	This file contains the yacc/bison grammar and semantic actions for the
 * mpl compiler.
 *
 * Ed Loyot 7/12/90
 *
 *              Copyright (c) 1990, 1991
 *              University of Virginia
 *              All Rights Reserved
 *
 *		This software is the property of the University of Virginia.
 *		Permission is granted to copy and distribute this software as
 *		long as this notice is intact on all copies.  Derived works 
 *  		must retain this notice.  This software is provided `as is'
 *		and without any express or implied warrenties.
 *
 *              Send problems & suggestions to:
 *		Ed Loyot (ecl2v@virginia.edu) or
 *              Andrew Grimshaw (grimshaw@cs.virginia.edu)
 *
 *****************************************************************************/

%{

    /* Copyright (C) 1989,1990 James A. Roskind, All rights reserved.
    This grammar was developed  and  written  by  James  A.  Roskind.
    Copying  of  this  grammar  description, as a whole, is permitted
    providing this notice is intact and applicable  in  all  complete
    copies.   Translations as a whole to other parser generator input
    languages  (or  grammar  description  languages)   is   permitted
    provided  that  this  notice is intact and applicable in all such
    copies,  along  with  a  disclaimer  that  the  contents  are   a
    translation.   The reproduction of derived text, such as modified
    versions of this grammar, or the output of parser generators,  is
    permitted,  provided  the  resulting  work includes the copyright
    notice "Portions Copyright (c)  1989,  1990  James  A.  Roskind".
    Derived products, such as compilers, translators, browsers, etc.,
    that  use  this  grammar,  must also provide the notice "Portions
    Copyright  (c)  1989,  1990  James  A.  Roskind"  in   a   manner
    appropriate  to  the  utility,  and in keeping with copyright law
    (e.g.: EITHER displayed when first invoked/executed; OR displayed
    continuously on display terminal; OR via placement in the  object
    code  in  form  readable in a printout, with or near the title of
    the work, or at the end of the file).  No royalties, licenses  or
    commissions  of  any  kind are required to copy this grammar, its
    translations, or derivative products, when the copies are made in
    compliance with this notice. Persons or corporations that do make
    copies in compliance with this notice may charge  whatever  price
    is  agreeable  to  a  buyer, for such copies or derivative works.
    THIS GRAMMAR IS PROVIDED ``AS IS'' AND  WITHOUT  ANY  EXPRESS  OR
    IMPLIED  WARRANTIES,  INCLUDING,  WITHOUT LIMITATION, THE IMPLIED
    WARRANTIES  OF  MERCHANTABILITY  AND  FITNESS  FOR  A  PARTICULAR
    PURPOSE.

    James A. Roskind
    Independent Consultant
    516 Latania Palm Drive
    Indialantic FL, 32903
    (407)729-4348
    jar@ileaf.com
    or ...!uunet!leafusa!jar

    ---end of copyright notice---*/

/******************************************************************************
 *
 *				 Includes and Defines 
 *
 *****************************************************************************/

#ifdef SABER_CC
#include <libc.h>
#include <malloc.h>
#endif

#include <stream.h>
#include <string.h>

#include "syntax_tree.h"
#include "symbol_table.h"
#include "tokens.h"
#include "mplfront.h"

#ifdef DEBUG_PARSER
#define EXECUTE_DEBUG(X) X
#else
#define EXECUTE_DEBUG(X)
#endif

#ifdef DEBUG_SYMBOL_TABLES
#define TABLE_DEBUG(X) X
#else
#define TABLE_DEBUG(X)
#endif

extern "C" int __builtin_alloca(...);

extern int 		   yylex();
extern int 		   column;
extern int 		   linenum;
extern void 		   yyerror(char *);
extern char 		   program_name[];
extern identifier 	   *id_redeclared_error(identifier *);
extern identifier 	   *id_redeclared_error(char *);
extern symbol_table 	   global_table;
extern symbol_table_stack  scope_stack;
extern symbol_table 	   *parameter_table;
extern definition_list     *parser_return_value;
extern class_definition    *mentat_class;
extern function_definition *main_function;
extern void 		   general_declaration(type *, declarator *, 
					       TOKEN = NONE, boolean = FALSE);

char 		*op_to_string(TOKEN);
boolean		match_token(TOKEN, TOKEN);
class_def_stack current_class;
TOKEN  		current_access_token;
boolean		main_flag = FALSE;

overload     	id_redeclared_error;

%}

/******************************************************************************
 *
 *				 Parser Stack Types
 *
 *****************************************************************************/

%union {
    int		          integer;
    TOKEN	          token;
    char 	          *string;
    identifier            *id_ptr;
    symbol_table          *table_ptr;
    expression            *expr_ptr;
    type		  *type_ptr;
    base_type             *base_type_ptr;
    type_modifier	  *type_modifier_ptr;
    storage_class	  *storage_class_ptr;
    parameter		  *parameter_ptr;
    parameter_list	  *parameter_list_ptr;
    declarator            *declarator_ptr;
    declarator_list       *declarator_list_ptr;
    initializer           *initializer_ptr;
    initializer_list      *initializer_list_ptr;
    declaration	          *declaration_ptr;
    statement	          *statement_ptr;
    statement_list        *statement_list_ptr;
    expression            *expression_ptr;
    expression_list       *expression_list_ptr;
    string_list           *string_list_ptr;
    definition            *definition_ptr;
    function_definition   *function_definition_ptr;
    definition_list       *definition_list_ptr;
    class_definition      *class_definition_ptr;
    enum_definition       *enum_definition_ptr;
    variable_declaration  *variable_declaration_ptr;
    constant		  *constant_ptr;
    c_string_list	  *c_string_list_ptr;
    ctor_initializer      *ctor_initializer_ptr;
    ctor_initializer_list *ctor_initializer_list_ptr;
    derivation_list       *derivation_list_ptr;
};

/******************************************************************************
 *
 *				Token Declarations
 *
 *****************************************************************************/

%token <token> AUTO            DOUBLE          INT             STRUCT
%token <token> BREAK           ELSE            LONG            SWITCH
%token <token> CASE            ENUM            REGISTER        TYPEDEF
%token <token> CHAR            EXTERN          RETURN          UNION
%token <token> CONST           FLOAT           SHORT           UNSIGNED
%token <token> CONTINUE        FOR             SIGNED          VOID
%token <token> DEFAULT         GOTO            SIZEOF          VOLATILE
%token <token> DO              IF              STATIC          WHILE
%token <token> NEW             DELETE
%token <token> THIS            
%token <token> OPERATOR
%token <token> CLASS           
%token <token> PUBLIC          PROTECTED       PRIVATE
%token <token> VIRTUAL         FRIEND
%token <token> INLINE

%token <string> IDENTIFIER             STRINGliteral
%token <string> FLOATINGconstant       INTEGERconstant        CHARACTERconstant
%token <string> OCTALconstant          HEXconstant
%token <string> TYPEDEFname 

%token <token> ARROW             /*    ->                              */
%token <token> ICR DECR          /*    ++      --                      */
%token <token> LS RS             /*    <<      >>                      */
%token <token> LE GE EQ NE       /*    <=      >=      ==      !=      */
%token <token> ANDAND OROR       /*    &&      ||                      */
%token <token> ELLIPSIS          /*    ...                             */
%token <token> CLCL              /*    ::                              */
%token <token> DOTstar ARROWstar /*    .*       ->*                    */
%token <token> MULTassign DIVassign   MODassign /*   *=      /=      %=      */
%token <token> PLUSassign MINUSassign           /*   +=      -=              */
%token <token> LSassign   RSassign              /*   <<=     >>=             */
%token <token> ANDassign  ERassign     ORassign /*   &=      ^=      |=      */
%token <token> '+' '-' '*' '/' '%' '^' '&' '|' '~' '!' '<' '>' '.' '(' ')'
%token <token> '[' ']' ',' '='

%token <token> MENTAT PERSISTENT REGULAR  /* Keywords for mpl. */

/******************************************************************************
 *
 *				Non-terminal Type Declarations
 *
 *****************************************************************************/

%type <token> assignment.operator aggregate.key access.specifier 
%type <token> virtual.opt indirect.or.reference access.specifier.opt
%type <token> operator.new clcl.opt.delete any.operator
%type <token> member.function.definition.semi.opt
%type <string> identifier.or.typedef.name rescoped.identifier.or.typedef.name
%type <id_ptr> rescoped.identifier class.rescoped.identifier
%type <id_ptr> operator.function.name
%type <table_ptr> start.compound.statement
%type <type_ptr> array.abstract.declarator declaration.qualifier.list
%type <type_ptr> type.qualifier.list typedef.declaration.specifier
%type <type_ptr> type.qualifier.list.opt
%type <type_ptr> typedef.type.specifier declaration.specifier type.specifier
%type <type_ptr> basic.declaration.specifier basic.type.specifier
%type <type_ptr> pointer.operator postfix.abstract.declarator 
%type <type_ptr> unary.abstract.declarator postfixing.abstract.declarator 
%type <type_ptr> abstract.declarator type.specifier.or.name
%type <type_ptr> operator.new.declarator.opt operator.new.array.declarator
%type <type_ptr> operator.function.ptr.opt
%type <type_ptr> elaborated.type.name sue.type.specifier
%type <type_ptr> sue.declaration.specifier
%type <base_type_ptr> basic.type.name
%type <storage_class_ptr> storage.class
%type <type_modifier_ptr> declaration.qualifier type.qualifier
%type <parameter_ptr> comma.opt.ellipsis parameter.declaration
%type <parameter_ptr> named.parameter.declaration
%type <parameter_list_ptr> parameter.list named.parameter.type.list
%type <parameter_list_ptr> parameter.type.list void.opt.param.list
%type <declarator_ptr> paren.identifier.declarator identifier.declarator 
%type <declarator_ptr> declarator unary.identifier.declarator
%type <declarator_ptr> postfix.identifier.declarator constructed.declarator
%type <declarator_ptr> nonunary.constructed.identifier.declarator
%type <declarator_ptr> constructed.identifier.declarator
%type <declarator_ptr> constructed.parameter.typedef.declarator
%type <declarator_ptr> constructed.paren.typedef.declarator typedef.declarator
%type <declarator_ptr> parameter.typedef.declarator clean.typedef.declarator
%type <declarator_ptr> clean.postfix.typedef.declarator
%type <declarator_ptr> paren.postfix.typedef.declarator
%type <declarator_ptr> paren.typedef.declarator simple.paren.typedef.declarator
%type <declarator_ptr> bit.field.identifier.declarator bit.field.declarator
%type <declarator_ptr> operator.new.type
%type <declarator_list_ptr> enumerator.list.no.trailing.comma enumerator.list
%type <initializer_ptr> initializer.opt initializer initializer.group
%type <initializer_list_ptr> initializer.list
%type <variable_declaration_ptr> declaring.list member.declaring.list
%type <variable_declaration_ptr> default.declaring.list type.name
%type <variable_declaration_ptr> member.default.declaring.list
%type <variable_declaration_ptr> member.conflict.paren.declaring.item
%type <variable_declaration_ptr> member.conflict.paren.postfix.declaring.item
%type <variable_declaration_ptr> member.conflict.declaring.item
%type <declaration_ptr> declaration
%type <function_definition_ptr> new.function.definition destructor.function
%type <function_definition_ptr> constructor.function.in.class
%type <function_definition_ptr> constructor.parameter.list.and.body
%type <function_definition_ptr> constructor.conflicting.parameter.list.and.body
%type <function_definition_ptr> constructor.conflicting.typedef.declarator
%type <function_definition_ptr> function.definition function.declaration
%type <function_definition_ptr> constructor.function.definition
%type <definition_ptr> external.definition
%type <definition_list_ptr> translation.unit prog.start
%type <statement_ptr> statement compound.statement expression.statement
%type <statement_ptr> selection.statement iteration.statement jump.statement
%type <statement_ptr> labeled.statement member.declaration
%type <statement_ptr> destructor.function.body
%type <statement_list_ptr> statement.list declaration.list
%type <statement_list_ptr> member.declaration.list.opt
%type <constant_ptr> linkage.specifier
%type <expression_ptr> constant primary.expression postfix.expression
%type <expression_ptr> unary.expression cast.expression deallocation.expression
%type <expression_ptr> point.member.expression multiplicative.expression
%type <expression_ptr> additive.expression shift.expression
%type <expression_ptr> relational.expression equality.expression AND.expression
%type <expression_ptr> exclusive.OR.expression inclusive.OR.expression
%type <expression_ptr> logical.AND.expression logical.OR.expression
%type <expression_ptr> conditional.expression assignment.expression
%type <expression_ptr> expression expression.opt constant.expression
%type <expression_ptr> member.pure.opt allocation.expression
%type <expression_ptr> enumerator.value.opt
%type <expression_ptr> string.literal.list
%type <expression_list_ptr> argument.expression.list
%type <expression_list_ptr> operator.new.initializer.opt 
%type <class_definition_ptr> aggregate.name
%type <enum_definition_ptr> enum.name
%type <derivation_list_ptr> parent.class derivation.list derivation.opt
%type <ctor_initializer_ptr> constructor.init
%type <ctor_initializer_list_ptr> constructor.init.list 
%type <ctor_initializer_list_ptr> constructor.init.list.opt

/******************************************************************************
 *
 *				Start Symbol Declaration
 *
 *****************************************************************************/

%start prog.start

/******************************************************************************
 *
 *				Grammar
 *
 *****************************************************************************/

%%
prog.start:
        /*nothing*/
            {
		// Parse the empty program.
		parser_return_value = NULL;
	    }
        | translation.unit
            {
		parser_return_value = $1;
		TABLE_DEBUG(scope_stack.dump_top();)
	    }
        ;

/******************************************************************************
 *
 *				 Constants
 *
 *****************************************************************************/

constant:
        /*  
	 * We are not including ENUMERATIONconstant here because we
         * are treating it like a variable with a type of "enumeration
         * constant".
	 */
        FLOATINGconstant
            {    
		EXECUTE_DEBUG(cerr << "Floating constant : " << $1 << "!\n";)
		$$ = new constant($1, FLOAT, FLOATINGconstant);
	    }
        | INTEGERconstant
            {
		EXECUTE_DEBUG(cerr << "Integer constant : " << $1 << "!\n";)
		$$ = new constant($1, INT, INTEGERconstant);
	    }
        | OCTALconstant
            {    
		EXECUTE_DEBUG(cerr << "Octal constant : " << $1 << "!\n";)
		$$ = new constant($1, INT, OCTALconstant);
	    }
        | HEXconstant
            {    
		EXECUTE_DEBUG(cerr << "Hex constant : " << $1 << "!\n";)
		$$ = new constant($1, INT, HEXconstant);
	    }
        | CHARACTERconstant
            {    
		EXECUTE_DEBUG(cerr << "Character constant : " << $1 << "!\n";)
		$$ = new constant($1, CHAR, CHARACTERconstant);
	    }
        ;

/******************************************************************************
 *
 * 				String Literals
 *
 *****************************************************************************/

string.literal.list:
                STRINGliteral
                    {
			constant *constant_ptr = new constant($1, 
						 STRINGliteral, STRINGliteral);
			$$ = new string_list(constant_ptr);
		    }
                | string.literal.list STRINGliteral
                    {
			constant *constant_ptr = new constant($2, 
						 STRINGliteral, STRINGliteral);
			((string_list *)$1)->add_string(constant_ptr);
			$$ = $1;
		    }    
                ;


/******************************************************************************
 *
 *				Expressions
 *
 *****************************************************************************/

/*
 * The following can be used in an identifier based declarator.  
 * (Declarators   that  redefine  an  existing  TYPEDEFname  require 
 * special handling, and are not included here).  In  addition,  the 
 * following are all valid "identifiers" in an expression, whereas a 
 * TYPEDEFname is NOT.
 */
    
rescoped.identifier:
        IDENTIFIER  /* ANSI C: We cannot use a TYPEDEFname as a variable */
            {
		$$ = new identifier($1, IDENTIFIER);
	    }
        | operator.function.name /* C++, not ANSI C*/
        | class.rescoped.identifier
        ;


/* 
 * When nested types are supported, an action must be taken immediatly
 * after the  CLCL  in each of the following productions to notify the 
 * lexer of a temporary change in scope. This change will allow  the 
 * lexer to deduce whether the name that follows the CLCL is that of 
 * an  IDENTIFIER  (member),  or  of a TYPEDEFname (locally declared 
 * type).  This lexical distinction is critical to continued  syntax 
 * analysis. 
 */

class.rescoped.identifier:
        TYPEDEFname CLCL { } identifier.or.typedef.name /* C++, not ANSI C */
            {
		/*
		 * The identifier.or.typedef.name will be an identifier if this
		 * is member or a TYPEDEFname if this is a nested type.
		 */

		$$ = new identifier($4, IDENTIFIER);

		/*
		 * Lookup the TYPEDEFname, which is a class name, in the symbol
		 * table.  It must be in the symbol table for the lexer to have
		 * identified this as a typedef name, so no need to check the 
		 * return value.  Get the class names class definition and use
		 * it to build a derivation list.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);

		class_definition *class_definition_ptr = 
		    typedef_id_ptr->csue_definition_ptr;
		if (class_definition_ptr == NULL)
		{
		    /* Log an error, a non class or forward defined class is 
		     * being used in a derivation list.
		     * - ecl. fix this later  and also in derivation.list:*/
		}
		$$->scope_list = new derivation_list(class_definition_ptr);
	    }
        | TYPEDEFname CLCL { } operator.function.name /* C++, not ANSI C */
            {
		$$ = $4;
		/*
		 * Lookup the TYPEDEFname, which is a class name, in the symbol
		 * table.  It must be in the symbol table for the lexer to have
		 * identified this as a typedef name, so no need to check the 
		 * return value.  Get the class names class definition and use
		 * it to build a derivation list.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);

		class_definition *class_definition_ptr = 
		    typedef_id_ptr->csue_definition_ptr;
		if (class_definition_ptr == NULL)
		{
		    /* Log an error, a non class or forward defined class is 
		     * being used in a derivation list.
		     * - ecl. fix this later  and also in derivation.list:*/
		}
		$$->scope_list = new derivation_list(class_definition_ptr);
	    }
        | TYPEDEFname CLCL { } '~' TYPEDEFname /* C++, not ANSI C */
            {
		$$ = new identifier($5, TYPEDEFname);
		/*
		 * Lookup the TYPEDEFname, which is a class name, in the symbol
		 * table.  It must be in the symbol table for the lexer to have
		 * identified this as a typedef name, so no need to check the 
		 * return value.  Get the class names class definition and use
		 * it to build a derivation list.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);

		class_definition *class_definition_ptr = 
		    typedef_id_ptr->csue_definition_ptr;
		if (class_definition_ptr == NULL)
		{
		    /* Log an error, a non class or forward defined class is 
		     * being used in a derivation list.
		     * - ecl. fix this later  and also in derivation.list:*/
		}
		$$->scope_list = new derivation_list(class_definition_ptr);
		$$->destructor_flag = TRUE;
	    }
        | TYPEDEFname CLCL { } class.rescoped.identifier /* C++, not ANSI C */
            {
		/*
		 * Lookup the TYPEDEFname, which is a class name, in the symbol
		 * table.  It must be in the symbol table for the lexer to have
		 * identified this as a typedef name, so no need to check the 
		 * return value.  Get the class names class definition and use
		 * it to build a derivation list.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);

		class_definition *class_definition_ptr = 
		    typedef_id_ptr->csue_definition_ptr;
		if (class_definition_ptr == NULL)
		{
		    /* Log an error, a non class or forward defined class is 
		     * being used in a derivation list.
		     * - ecl. fix this later  and also in derivation.list:*/
		}
		$4->scope_list->add_derivation_front(class_definition_ptr);
		$$ = $4;
	    }
        ;

/*
 * Note that the following MUST come before primary.expression so 
 * that  the  reduce/reduce conflict is properly resolved (according 
 * to Stroustrup: "If it can be a declaration, then it  should  be".  
 * I  just  give  up  and  go  in  that  direction  after  I see the 
 * declarator, whereas cfront tries to lex  ahead  to  disambiguate.  
 * My reduce-reduce conflict disambiguates this situation at a point 
 * where human readers cannot generally disambiguate (mentally). 
 */
    
paren.identifier.declarator:
        rescoped.identifier
	    {
		// The identifier in this rule will be used in a declaration.
		$$ = new declarator($1);
	    }
        | '(' paren.identifier.declarator ')' 
	    { 
		// Ignore the parenthesis for now.
		$$ = $2; 
	    }
        ;

/*
 * Note that CLCL IDENTIFIER is NOT part of rescoped.identifier. 
 * It is ONLY valid for referring to an identifier,  and  NOT  valid 
 * for  declaring  (or  importing  an  external  declaration  of) an 
 * identifier.  This disambiguates the following code,  which  would 
 * otherwise be syntactically and semantically ambiguous:
 *
 *          class base {
 *              static int i; // element i;
 *              float member_funct(void);
 *              };
 *          base i; // global i
 *          float base::member_function(void) {
 *              i; // refers to static int element "i" of base
 *              ::i; // refers to global "i", with type "base"
 *                  {
 *                  base :: i; // import of global "i", like "base (::i);"?
 *                              // OR reference to global??
 *                  }
 *              }
 */

primary.expression:
        CLCL identifier.or.typedef.name
            {
		/*
		 * Need to lookup  the identifier.or.typedef.name in the global
		 * scope.  If it is not found, log an error and enter the
		 * identifier in the global symbol table to continue the parse.
		 */

		identifier *id_ptr = global_table.lookup($2);
		if (id_ptr == NULL)
		{
		    ostream error(max_message_size, message_buffer);
		    error << "Undefined identifier or typedef name: " << $2;
		    error << "used in expression!\n";
		    error.put('\0');
		    yyerror(message_buffer);
		    id_ptr = global_table.insert($2, IDENTIFIER);
		    id_ptr->type_ptr = new type(new base_type(INT));
		}

		$$ = new binary_expression(NULL, id_ptr, CLCL, 
					   id_ptr->type_ptr->duplicate());
	    }
        | CLCL operator.function.name 
            {
		/*
		 * Need to lookup operator.function.name in the global
		 * scope.  If it is not found, log an error and enter the
		 * identifier to continue the parse.
		 */

		identifier *id_ptr = global_table.lookup($2->id_string);
		if (id_ptr == NULL)
		{
		    ostream error(max_message_size, message_buffer);
		    error << "Undefined operator function name : ";
		    error << $2->id_string << "used in expression!\n";
		    error.put('\0');
		    yyerror(message_buffer);
		    id_ptr = global_table.insert($2);
		    id_ptr->operator_flag = TRUE;
		    id_ptr->operator_token = OPERATOR;
		    id_ptr->type_ptr = new type(new base_type(FUNCTION));
		}

		$$ = id_ptr;
	    }
        | THIS
            {
		// Get the type for THIS from the current class.
		type *this_type_ptr = NULL;
		
		if (current_class.top != NULL)
		{
		    // Get the class name.
		    identifier *id_ptr = 
			current_class.top->class_def_ptr->identifier_ptr;

		    // Get the classes type.
		    type *temp_type_ptr = id_ptr->type_ptr->duplicate();
		    temp_type_ptr->expand_typedefs();

		    // This is a pointer so build the type, pointer to class.
		    this_type_ptr = new type(new type_modifier('*'));
		    this_type_ptr->add_type(temp_type_ptr); 
		}
		    
		$$ = new binary_expression(NULL, NULL, THIS, this_type_ptr);
	    }
        | rescoped.identifier
            {
		/*
		 * An identifier is being referenced in an expression. Lookup
		 * the identifier in the symbol table.  If it is not found
		 * log the error and enter the identifier to continue the 
		 * parse.
		 */
		
		identifier *id_ptr = NULL;
		
		/*
		 * If the identifier was modified with scoping information
		 * use this information to build a scoping expression and 
		 * lookup the identifier in the proper classes' scope.
		 */
		
		expression *scope_expr_ptr = NULL;

		if ($1->scope_list != NULL)
		{
		    scope_expr_ptr = $1->scope_list->build_scope_expression();
		    symbol_table_stack *scope_ptr = 
			$1->scope_list->get_scope();
		    if (scope_ptr != NULL)
			id_ptr = scope_ptr->lookup($1->id_string);
		}
		else
		    if (parameter_table != NULL)
		    {
			id_ptr = parameter_table->lookup($1->id_string);
			if (id_ptr == NULL)
			    id_ptr = scope_stack.lookup($1->id_string);
		    }
		    else
			id_ptr = scope_stack.lookup($1->id_string);

		if (id_ptr == NULL)
		{
		    ostream error(max_message_size, message_buffer);
		    error << "Undefined identifier : " << $1->id_string;
		    error << " used in expression!\n";
		    error.put('\0');
		    yyerror(message_buffer);
		    if (parameter_table != NULL)
			id_ptr = parameter_table->insert($1);
		    else
			id_ptr = scope_stack.insert($1);
		    id_ptr->type_ptr = new type(new base_type(INT));
		}
		else
		    // Delete the old identifier record.
		    delete $1;

		if (scope_expr_ptr != NULL)
		{
		    type *type_ptr = id_ptr->type_ptr->duplicate();
		    $$ = new binary_expression(scope_expr_ptr, id_ptr, CLCL,
					       type_ptr);
		}
		else
		    $$ = id_ptr;
	    }
        | constant
        | string.literal.list
        | '(' expression ')'
            {
		$$ = new expression_list($2, $2->type_ptr->duplicate());
	    }
        ;

/*
 * The  following  introduces  MANY  conflicts.   Requiring  and 
 * allowing '(' ')' around the `type' when the type is complex would 
 * help a lot. 
 */

operator.function.name:
        OPERATOR any.operator
	    {
		// This rule covers overloaded operators.
		$$ = new identifier(op_to_string($2), OPERATOR);
		$$->operator_flag = TRUE;
		$$->operator_token = $2;
	    }
        | OPERATOR type.specifier.or.name  operator.function.ptr.opt
	    {
		// This rule covers conversion functions E&S 12.3.2.
		$$ = new identifier(NULL, OPERATOR);
		$$->operator_flag = TRUE;
		$2->add_type($3);
		$$->conversion_type_ptr = $2;
	    }
        | OPERATOR type.qualifier.list     operator.function.ptr.opt
	    {
		// This rule covers conversion functions E&S 12.3.2.
		$$ = new identifier(NULL, OPERATOR);
		$$->operator_flag = TRUE;
		$2->add_type($3);
		$$->conversion_type_ptr = $2;
	    }
        ;

/*
 * The following causes several ambiguities on *  and  &.   These 
 * conflicts  would also be removed if parens around the `type' were 
 * required in the derivations for operator.function.name 
 *
 * Interesting  aside:  The  use  of  right  recursion  in   the 
 * production  for  operator.function.ptr.opt gives both the correct 
 * parsing, AND removes a conflict!   Right  recursion  permits  the 
 * parser   to   defer   reductions  (aka:  delay  resolution),  and 
 * effectively make a second pass! 
 */
        
operator.function.ptr.opt:
        /* nothing */
            {
		$$ = NULL;
	    }
        | pointer.operator operator.function.ptr.opt
            {
		if ($2 == NULL)
		    $$ = $1;
		else
		{
		    $2->add_type($1);
		    $$ = $2;
		}
	    }
        | indirect.or.reference operator.function.ptr.opt
            {
		if ($2 == NULL)
		    $$ = new type(new type_modifier($1));
		else
		{
		    $$ = $2;
		    $$->add_type_element(new type_modifier($1));
		}
	    }
        ;

any.operator:
        '+'
        | '-'
        | '*'
        | '/'
        | '%'
        | '^'
        | '&'
        | '|'
        | '~'
        | '!'
        | '<'
        | '>'
        | LS
        | RS
        | ANDAND
        | OROR
        | ARROW
        | ARROWstar

/* 
 * These cannot be overloaded.
 *      | '.' 
 *      | DOTstar
 */

        | ICR
        | DECR
        | LE
        | GE
        | EQ
        | NE
        | assignment.operator
        | '(' ')'
        | '[' ']'
        | NEW
        | DELETE
        | ','
        ;

/*
 * The following production for type.qualifier.list was specially 
 * placed BEFORE the definition of postfix.expression to  resolve  a 
 * reduce-reduce conflict correctly 
 */        

type.qualifier.list.opt:
        /* Nothing */
            {
		$$ = NULL;
	    }
        | type.qualifier.list
        ;

/*
 * Note  that  the next set of productions in this grammar gives 
 * post-increment a higher precedence that pre-increment.   This  is 
 * not  clearly  stated  in  the  C++  Reference manual, and is only 
 * implied by the grammar in the ANSI C Standard. 
 */

postfix.expression:
        primary.expression
        | postfix.expression '[' expression ']'
            {
		/*
		 * Make sure the postfix.expression is an array.  If not log an
		 * error.  Set the type of this expression to be the base type
		 * for the array.
		 *
		 * This could be a class with an overloaded operator, must
		 * check for this also.  Add this later - ecl.
		 */

		if ($1->type_ptr == NULL || !$1->type_ptr->array_type())
		{
		    /*
		     * The postfix expression is not an array. Log the error.
		     * Fix this later - ecl.
		     */
		}
		
		type *type_ptr = $1->type_ptr->get_array_base_type();
		
		$$ = new binary_expression($1, $3, '[', type_ptr);
	    }
        | postfix.expression '(' ')'
            {
		/*
		 * Make sure the postfix.expression is a function.  If not log
		 * an error.  Set the type of this expression to be the return
		 * type	for the function.
		 *
		 * This could be a class with an overloaded operator, must
		 * check for this also.  Add this later - ecl.
		 */

		type *type_ptr;

		if ($1->type_ptr == NULL || !$1->type_ptr->function_type())
		{
		    /*
		     * The postfix expression is not a function. Log the error.
		     * Fix this later - ecl.
		     */
		    
		    type_ptr = $1->type_ptr;
		}
		else
		    type_ptr = $1->type_ptr->get_function_return_type();
		
		$$ = new binary_expression($1, new expression_list(NULL), '(',
					    type_ptr);
	    }
        | postfix.expression '(' argument.expression.list ')'
            {
		/*
		 * Make sure the postfix.expression is a function.  If not log
		 * an error.  Set the type of this expression to be the return
		 * type	for the function.
		 *
		 * This could be a class with an overloaded operator, must
		 * check for this also.  Add this later - ecl.
		 */

		type *type_ptr;
		
		if ($1->type_ptr == NULL || !$1->type_ptr->function_type())
		{
		    /*
		     * The postfix expression is not a function. Log the error.
		     * Fix this later - ecl.
		     */
		    
		    type_ptr = $1->type_ptr;
		}
		else
		    type_ptr = $1->type_ptr->get_function_return_type();
		
		$$ = new binary_expression($1,(expression *)$3, '(', type_ptr);
	    }
        | postfix.expression '.' rescoped.identifier.or.typedef.name
            {
		/*
		 * The postfix.expression must be a class object. Lookup the 
		 * rescoped.identifier.or.typedef.name string in the class
		 * definition.
		 */

		identifier *id_ptr;
		if ($1->type_ptr->class_object_type())
		{
		    class_definition *class_def_ptr = 
			$1->type_ptr->get_class_definition();
		    id_ptr = class_def_ptr->lookup($3);
		    if (id_ptr == NULL)
		    {
			//log an error if the identifier is undefined. - ecl.
			id_ptr = new identifier($3, IDENTIFIER);
			id_ptr->type_ptr = new type(new base_type(INT));
		    }
		}
		else
		{
		    // Log an error if this is not a class object. - ecl.
		    id_ptr = new identifier($3, IDENTIFIER);
		    id_ptr->type_ptr = new type(new base_type(INT));
		}

		type *type_ptr = id_ptr->type_ptr->duplicate();
	        $$ = new binary_expression($1, id_ptr, '.', type_ptr);
	    }
        | postfix.expression ARROW rescoped.identifier.or.typedef.name
            {
		/*
		 * The postfix.expression must be a pointer to a class object. 
		 * Lookup the  rescoped.identifier.or.typedef.name string in 
		 * the class definition.
		 */

		identifier *id_ptr;
		type *expression_type_ptr=$1->type_ptr->get_dereference_type();
		if (expression_type_ptr != NULL &&
		    expression_type_ptr->class_object_type())
		{
		    class_definition *class_def_ptr = 
			expression_type_ptr->get_class_definition();
		    id_ptr = class_def_ptr->lookup($3);
		    if (id_ptr == NULL)
		    {
			//log an error if the identifier is undefined. - ecl.
			id_ptr = new identifier($3, IDENTIFIER);
			id_ptr->type_ptr = new type(new base_type(INT));
		    }
		}
		else
		{
		    // Log an error if this is not a class object pointer -ecl.
		    id_ptr = new identifier($3, IDENTIFIER);
		    id_ptr->type_ptr = new type(new base_type(INT));
		}

		type *type_ptr = id_ptr->type_ptr->duplicate();
	        $$ = new binary_expression($1, id_ptr, ARROW, type_ptr);
	    }
        | postfix.expression ICR
            {
		$$ = new binary_expression($1, NULL, ICR, $1->type_ptr);
	    }
        | postfix.expression DECR
            {
		$$ = new binary_expression($1, NULL, DECR, $1->type_ptr);
	    }

                /* The next 4 rules are the source of cast ambiguity */
        | TYPEDEFname '(' ')'
            {
		/*
		 * This construct is called an explicit type conversion.  Its
		 * effect is to create an object of the specified type.  In
		 * other words it is similar to new. (see E&S p52)
		 */

		identifier *id_ptr = scope_stack.lookup($1);
		
		type *type_ptr = id_ptr->type_ptr->duplicate();
		type_ptr->expand_typedefs();
		$$ = new binary_expression(id_ptr, new expression_list(NULL),
					   '(', type_ptr);
	    }
        | TYPEDEFname '(' argument.expression.list ')'
            {
		/*
		 * This construct is called an explicit type conversion.  Its
		 * effect is to create an object of the specified type.  In
		 * other words it is similar to new. (see E&S p52)
		 */

		identifier *id_ptr = scope_stack.lookup($1);
		
		type *type_ptr = id_ptr->type_ptr->duplicate();
		type_ptr->expand_typedefs();
		$$ = new binary_expression(id_ptr, (expression *)$3,
					   '(', type_ptr);
	    }
        | basic.type.name '(' assignment.expression ')'
            {
		/*
		 * This construct is called an explicit type conversion.  Its
		 * effect is to create an object of the specified type.  In
		 * other words it does the same thing as new does.
		 */

		type *type_ptr = new type($1);

		type_expression *type_expression_ptr = 
		    new type_expression(new variable_declaration(type_ptr));

		identifier *id_ptr = 
		    new identifier($1->get_string(), IDENTIFIER);

		$$ = new binary_expression(id_ptr, new expression_list($3), 
					   '(', type_ptr);
	    }


        /* 
	 * If the following rule is added to the grammar, there will
	 * be 3 additional reduce-reduce conflicts.  They will all be
         * resolved in favor of NOT using the following rule,so no harm
	 * will be done.  However, since the rule is semantically
         * illegal we will omit it until we are enhancing the grammar
         * for error recovery 
	 *
	 * In fact this production is legal (see example E&S p 53.)
	 * The grammar must be fixed.  - ecl.
	 */

/*      | basic.type.name '(' ')'  Illegal: no such constructor */
        ;

rescoped.identifier.or.typedef.name:
        TYPEDEFname 
        | rescoped.identifier
	    {
		$$ = $1->id_string;
		delete $1;
	    }
        ;

argument.expression.list:
        assignment.expression
            {
		$$ = new expression_list($1);
	    }
        | argument.expression.list ',' assignment.expression
            {
		$1->add_expression($3);
		$$ = $1;
	    }
        ;

unary.expression:
        postfix.expression
        | ICR unary.expression
            {
		$$ = new binary_expression(NULL, $2, ICR, $2->type_ptr);
	    }
        | DECR unary.expression
            {
		$$ = new binary_expression(NULL, $2, DECR, $2->type_ptr);
	    }
        | indirect.or.reference cast.expression
            {
		// Build the appropriate type.
		type *type_ptr;
		char operator_symbol;
		if ($1 == '*')
		{
		    operator_symbol = '*';
		    // Make sure the cast.expression is a pointer.
		    if ($2->type_ptr->pointer_type())
			type_ptr = $2->type_ptr->get_dereference_type();
		    else
			// Log an error. - ecl
			;
		}
		else
		{
		    operator_symbol = '&';
		    type_ptr = new type(new type_modifier('*'));
		    type_ptr->add_type($2->type_ptr);
		}
				
		$$ = new binary_expression(NULL, $2, operator_symbol,type_ptr);
	    }
        | '-' cast.expression
            {
		$$ = new binary_expression(NULL, $2, '-', 
					   $2->type_ptr->promote_integer());
	    }
        | '+' cast.expression
            {
		$$ = new binary_expression(NULL, $2, '+', 
					   $2->type_ptr->promote_integer());
	    }
        | '~' cast.expression
            {
		$$ = new binary_expression(NULL, $2, '~', 
					   $2->type_ptr->promote_integer());
	    }
        | '!' cast.expression
            {
		type *type_ptr = new type(new base_type(INT));
		$$ = new binary_expression(NULL, $2, '!', type_ptr);
	    }
        | SIZEOF unary.expression
            {
		// Sizeof returns an int. See E&S p 58 for correct type - ecl.
		type *type_ptr = new type(new base_type(INT));
		$$ = new binary_expression(NULL, $2, SIZEOF, type_ptr);
	    }
        | SIZEOF '(' type.name ')'
            {
		// Sizeof returns an int. See E&S p 58 for correct type - ecl.
		type *type_ptr = new type(new base_type(INT));
		type_expression *type_expr = new type_expression($3);
		expression_list *parens_expr = new expression_list(type_expr);
		$$ = new binary_expression(NULL, parens_expr, SIZEOF, 
					   type_ptr);
	    }
        | allocation.expression
        ;


/*
 * Note that I could have moved the  newstore  productions  to  a 
 * lower  precedence  level  than  multiplication  (binary '*'), and 
 * lower than bitwise AND (binary '&').  These moves  are  the  nice 
 * way  to  disambiguate a trailing unary '*' or '&' at the end of a 
 * freestore expression.  Since the freestore expression (with  such 
 * a  grammar  and  hence  precedence  given)  can never be the left 
 * operand of a binary '*' or '&', the ambiguity would  be  removed.  
 * These  problems  really  surface when the binary operators '*' or 
 * '&' are overloaded, but this must be syntactically  disambiguated 
 * before the semantic checking is performed...  Unfortunately, I am 
 * not  creating  the language, only writing a grammar that reflects 
 * its specification, and  hence  I  cannot  change  its  precedence 
 * assignments.   If  I  had  my  druthers,  I would probably prefer 
 * surrounding the type with parens all the time, and  avoiding  the 
 * dangling * and & problem all together.
 */

       /* Following are C++, not ANSI C */
allocation.expression:
        operator.new                                  '(' type.name ')'
                operator.new.initializer.opt
            {
		type_expression *type_expr = new type_expression($3);
		expression_list *expr_list = new expression_list(type_expr);
		$$ = new allocation_expression($1, expr_list);
		((allocation_expression *)$$)->initializers = $5;
	    }
        | operator.new '(' argument.expression.list ')' '(' type.name ')'
                operator.new.initializer.opt
            {
		type_expression *type_expr = new type_expression($6);
		expression_list *expr_list = new expression_list(type_expr);
		$$ = new allocation_expression($1, expr_list);
		((allocation_expression *)$$)->argument_list = $3;
		((allocation_expression *)$$)->initializers = $8;
	    }

        /* next two rules are the source of * and & ambiguities */
        | operator.new                                  operator.new.type
            {
		$$ = new allocation_expression($1, $2);
	    }
        | operator.new '(' argument.expression.list ')' operator.new.type
            {
		$$ = new allocation_expression($1, $5);
		((allocation_expression *)$$)->argument_list = $3;
	    }
        ;

operator.new:
        NEW
        | CLCL NEW
        ;

/*
 * Note:  the  otherwise  "silly" inline expansion in the next 4 
 * productions is necessary to allow such expressions as:
 *
 *    new const T :: * ;
 *    new const T ;
 *
 * to be parsed correctly.  
 */

operator.new.type:
        type.qualifier.list      operator.new.declarator.opt
                        operator.new.initializer.opt
            {
		type *type_ptr = $1->duplicate();
		if ($2 != NULL)
		{
		    $2->add_type(type_ptr);
		    $$ = new declarator($2);
		}
		else
		    $$ = new declarator(type_ptr);
		$$->parenthesized_initializers = $3;
	    }
        | type.specifier.or.name operator.new.declarator.opt
                        operator.new.initializer.opt
            {
		type *type_ptr = $1->duplicate();
		if ($2 != NULL)
		{
		    $2->add_type(type_ptr);
		    $$ = new declarator($2);
		}
		else
		    $$ = new declarator(type_ptr);
		$$->parenthesized_initializers = $3;
	    }
        ;

/* 
 * Perhaps I should move  the  following  to  a  more  consistent 
 * position, but for now I will just leave it here. 
 *
 *  Note that the following does not include type.qualifier.list. 
 * Hence, whenever type.specifier.or.name is used, an adjacent  rule 
 * is  supplied containing type.qualifier.list.  It is not generally 
 * possible    to    know    immediately    (i.e.,     reduce)     a 
 * type.qualifier.list,  as  a TYPEDEFname that follows might not be 
 * part of a type specifier, but might instead  be  "TYPEDEFname  :: 
 * *". 
 */

type.specifier.or.name:
        type.specifier
        | basic.type.name
            {
		$$ = new type($1);
	    }
        | TYPEDEFname
            {
		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		$$ = typedef_id_ptr->type_ptr->duplicate();
	    }
        ;

/*
 * Right  recursion  is critical in the following productions to 
 * avoid a conflict on TYPEDEFname 
 */

operator.new.declarator.opt:
        /* Nothing */
            {
		$$ = NULL;
	    }
        | operator.new.array.declarator
        | indirect.or.reference operator.new.declarator.opt
            {
		if ($2 != NULL)
		{
		    $2->add_type_element(new type_modifier($1));
		    $$ = $2;
		}
		else
		    $$ = new type(new type_modifier($1));
	    }
        | pointer.operator      operator.new.declarator.opt
            {
		if ($2 != NULL)
		{
		    $2->add_type($1);
		    $$ = $2;
		}
		else
		    $$ = $1;
	    }
        ;

operator.new.array.declarator:
        '[' ']'
            {
		$$ = new type(new type_modifier(ARRAY));
	    }
        | '[' expression ']'
            {
		$$ = new type(new type_modifier(ARRAY, $2));
	    }
        | operator.new.array.declarator '[' expression ']'
            {
		$1->add_type_element(new type_modifier(ARRAY, $3));
		$$ = $1;
	    }
        ;

operator.new.initializer.opt:
        /* Nothing */
            {
 		$$ = NULL;
	    }
        | '(' ')'
            {
		$$ = new expression_list(NULL);
	    }
        | '(' argument.expression.list ')'
            {
		$$ = $2;
	    }
        ;

cast.expression:
        unary.expression
        | '(' type.name ')' cast.expression
            {
		type_expression *type_expr = new type_expression($2);
		expression_list *parens_expr = new expression_list(type_expr);
		type *cast_type_ptr = $2->get_type()->duplicate();
		cast_type_ptr->expand_typedefs();
		$$ = new cast_expression(parens_expr, $4, cast_type_ptr);
	    }
        ;

deallocation.expression:
        cast.expression
        | clcl.opt.delete deallocation.expression
            {
		type *type_ptr = new type(new base_type(VOID));
		$$ = new deallocation_expression($1, $2, type_ptr);
	    }
        | clcl.opt.delete '[' expression ']' deallocation.expression
            {
		type *type_ptr = new type(new base_type(VOID));
		$$ = new deallocation_expression($1, $5, type_ptr);
		((deallocation_expression *)$$)->bracket_expression = 
		    new binary_expression(NULL, $3, '[');
	    }
        | clcl.opt.delete '[' ']' deallocation.expression
            {
		type *type_ptr = new type(new base_type(VOID));
		$$ = new deallocation_expression($1, $4, type_ptr);
		((deallocation_expression *)$$)->bracket_expression = 
		    new binary_expression(NULL, NULL, '[');
	    }
        ;

clcl.opt.delete:
        DELETE
        | CLCL DELETE
        ;

point.member.expression:
        deallocation.expression
        | point.member.expression DOTstar  deallocation.expression
            {
		$$ = new binary_expression($1, $3, $2);
	    }
        | point.member.expression ARROWstar  deallocation.expression
            {
		$$ = new binary_expression($1, $3, $2);
	    }
        ;
        
multiplicative.expression:
        point.member.expression
        | multiplicative.expression '*' point.member.expression
            {
		// Must consider overloaded operators. - ecl
		type *type_ptr;
		if ($1->type_ptr->arithmetic_type() &&
		    $3->type_ptr->arithmetic_type())
		    type_ptr = $1->type_ptr->convert_arithmetic($3->type_ptr);
		else
		    // Log an appropriate error. - ecl
		    ;
		$$ = new binary_expression($1, $3, '*', type_ptr);
	    }
        | multiplicative.expression '/' point.member.expression
            {
		// Must consider overloaded operators. - ecl
		type *type_ptr;
		if ($1->type_ptr->arithmetic_type() &&
		    $3->type_ptr->arithmetic_type())
		    type_ptr = $1->type_ptr->convert_arithmetic($3->type_ptr);
		else
		    // Log an appropriate error. - ecl
		    ;
		$$ = new binary_expression($1, $3, '/', type_ptr);
	    }
        | multiplicative.expression '%' point.member.expression
            {
		// Must consider overloaded operators. - ecl
		type *type_ptr;
		if ($1->type_ptr->integral_type() &&
		    $3->type_ptr->integral_type())
		    type_ptr = $1->type_ptr->convert_arithmetic($3->type_ptr);
		else
		    // Log an appropriate error. - ecl
		    ;
		$$ = new binary_expression($1, $3, '%', type_ptr);
	    }
        ;

additive.expression:
        multiplicative.expression
        | additive.expression '+' multiplicative.expression
            {
		// Must consider overloaded operators. - ecl
		type *type_ptr;
		if ($1->type_ptr->arithmetic_type() &&
		    $3->type_ptr->arithmetic_type())
		    type_ptr = $1->type_ptr->convert_arithmetic($3->type_ptr);
		else
		    if ($1->type_ptr->pointer_type() &&
			$3->type_ptr->integral_type())
			type_ptr = $1->type_ptr;
		    else
			// Log an appropriate error. - ecl.
			;
		
		$$ = new binary_expression($1, $3, '+', type_ptr);
	    }
        | additive.expression '-' multiplicative.expression
            {
		// Must consider overloaded operators. - ecl
		type *type_ptr;
		if ($1->type_ptr->arithmetic_type() &&
		    $3->type_ptr->arithmetic_type())
		    type_ptr = $1->type_ptr->convert_arithmetic($3->type_ptr);
		else
		    if ($1->type_ptr->pointer_type() &&
			$3->type_ptr->integral_type())
			type_ptr = $1->type_ptr;
		    else
			if ($1->type_ptr->pointer_type() &&
			    $3->type_ptr->pointer_type())
			    type_ptr = new type(new base_type(INT));
			else
			    // Log an appropriate error. - ecl.
			    ;
		
		$$ = new binary_expression($1, $3, '-', type_ptr);
	    }
        ;

shift.expression:
        additive.expression
        | shift.expression LS additive.expression
            {
		$$ = new binary_expression($1, $3, LS, $1->type_ptr);
	    }
        | shift.expression RS additive.expression
            {
		$$ = new binary_expression($1, $3, RS, $1->type_ptr);
	    }
        ;

relational.expression:
        shift.expression
        | relational.expression '<' shift.expression
            {
		type *type_ptr = new type(new base_type(INT));
		$$ = new binary_expression($1, $3, '<', type_ptr);
	    }
        | relational.expression '>' shift.expression
            {
		type *type_ptr = new type(new base_type(INT));
		$$ = new binary_expression($1, $3, '>', type_ptr);
	    }
        | relational.expression LE shift.expression
            {
		type *type_ptr = new type(new base_type(INT));
		$$ = new binary_expression($1, $3, LE, type_ptr);
	    }
        | relational.expression GE shift.expression
            {
		type *type_ptr = new type(new base_type(INT));
		$$ = new binary_expression($1, $3, GE, type_ptr);
	    }
        ;

equality.expression:
        relational.expression
        | equality.expression EQ relational.expression
            {
		type *type_ptr = new type(new base_type(INT));
		$$ = new binary_expression($1, $3, EQ, type_ptr);
	    }
        | equality.expression NE relational.expression
            {
		type *type_ptr = new type(new base_type(INT));
		$$ = new binary_expression($1, $3, NE, type_ptr);
	    }
        ;

AND.expression:
        equality.expression
        | AND.expression '&' equality.expression
            {
		type *type_ptr;
		if ($1->type_ptr->integral_type() &&
		    $3->type_ptr->integral_type())
		    type_ptr = $1->type_ptr->convert_arithmetic($3->type_ptr);
		else
		    // Log an appropriate error. - ecl
		    ;
		
		$$ = new binary_expression($1, $3, '&', type_ptr);
	    }
        ;

exclusive.OR.expression:
        AND.expression
        | exclusive.OR.expression '^' AND.expression
            {
		type *type_ptr;
		if ($1->type_ptr->integral_type() &&
		    $3->type_ptr->integral_type())
		    type_ptr = $1->type_ptr->convert_arithmetic($3->type_ptr);
		else
		    // Log an appropriate error. - ecl
		    ;

		$$ = new binary_expression($1, $3, '^', type_ptr);
	    }
        ;

inclusive.OR.expression:
        exclusive.OR.expression
        | inclusive.OR.expression '|' exclusive.OR.expression
            {
		type *type_ptr;
		if ($1->type_ptr->integral_type() &&
		    $3->type_ptr->integral_type())
		    type_ptr = $1->type_ptr->convert_arithmetic($3->type_ptr);
		else
		    // Log an appropriate error. - ecl
		    ;

		$$ = new binary_expression($1, $3, '|', type_ptr);
	    }
        ;

logical.AND.expression:
        inclusive.OR.expression
        | logical.AND.expression ANDAND inclusive.OR.expression
            {
		type *type_ptr = new type(new base_type(INT));
		$$ = new binary_expression($1, $3, ANDAND, type_ptr);
	    }
        ;

logical.OR.expression:
        logical.AND.expression
        | logical.OR.expression OROR logical.AND.expression
            {
		type *type_ptr = new type(new base_type(INT));
		$$ = new binary_expression($1, $3, OROR, type_ptr);
	    }
        ;

conditional.expression:
        logical.OR.expression
        | logical.OR.expression '?' expression ':'
                conditional.expression
            {
		expression *temp_expr_ptr = new binary_expression($3, $5, ':',
						    $3->type_ptr->duplicate());
		$$ = new binary_expression($1, temp_expr_ptr, '?', 
					   $3->type_ptr->duplicate());
	    }
        ;

assignment.expression:
        conditional.expression
        | unary.expression assignment.operator assignment.expression
            {
		EXECUTE_DEBUG(cerr << "Building assignment statement.\n";)
		$$ = new binary_expression($1, $3, $2, 
					   $1->type_ptr->duplicate());
	    }
        ;

assignment.operator:
        '='
        | MULTassign
        | DIVassign
        | MODassign
        | PLUSassign
        | MINUSassign
        | LSassign
        | RSassign
        | ANDassign
        | ERassign
        | ORassign
        ;

expression:
        assignment.expression
        | expression ',' assignment.expression
            {
		$$ = new binary_expression($1, $3, ',', $3->type_ptr);
	    }
        ;

constant.expression:
        conditional.expression
        ;

/* 
 * The following was used for clarity.
 */

expression.opt:
        /* Nothing */
            {
		$$ = NULL;
	    }
        | expression
        ;



/******************************************************************************
 *
 *				Declarations
 *
 *****************************************************************************/

/*
 * The following are notably different from the ANSI  C  Standard 
 * specified  grammar,  but  are  present  in  my  ANSI C compatible 
 * grammar.  The changes were made to disambiguate typedefs presence 
 * in   declaration.specifiers   (vs.    in   the   declarator   for 
 * redefinition);  to allow struct/union/enum/class tag declarations 
 * without  declarators,  and  to  better  reflect  the  parsing  of 
 * declarations     (declarators     must     be    combined    with 
 * declaration.specifiers ASAP, so that they can immediately  become 
 * visible in the current scope). 
 */

declaration:
        sue.declaration.specifier ';' 
            { 
		/* This is semantic error, as it includes a storage class!?! */
		/* 
		 * No can have the storage class typedef!, other are probably
		 * illegal. - ecl
		 *
		 * This must be fixed to handle typedefed structs like:
	         *
		 * typedef struct fred bob;
		 *
		 * It will also have to work like the case below.
		 * -ecl
		 */
		$$ = new csue_declaration($1);
	    }
        | sue.type.specifier ';'  /* Class etc. definition or declaration. */
            {
		/*
		 * This is the first place the parser can distinguish between
		 * the following situations:
		 *
		 * 1) struct struct_name *variable_name;
		 *
		 * 2) struct struct_name;  // Where struct_name has been
		 *			   // defined previously.
		 *
		 * Case 2 declares a type name, while case 1 requires a lookup.
		 * The variable redeclare_name_flag indicates that this is
		 * case 2.  The member function redeclare_name declares the
		 * name.
		 */

		identifier *id_ptr = $1->get_class_identifier();
		if (id_ptr != NULL && id_ptr->redeclare_name_flag)
		    // This could be an enum, thus the id_ptr would be NULL.
		    id_ptr->redeclare_name();
		    
		$$ = new csue_declaration($1);
	    }
        | declaring.list ';'
	    {
		$$ = $<declaration_ptr>1;
	    }
        | default.declaring.list ';'
	    {
		$$ = $<declaration_ptr>1;
	    }
        ;

/*
 * Note  that  if  a typedef were redeclared, then a declaration 
 * specifier must be supplied (re: ANSI C spec).  The following  are 
 * declarations  wherein  no  declaration.specifier is supplied, and 
 * hence the 'default' must be used.  An example of this is
 *
 *      const a;
 *
 * which by default, is the same as:
 *
 *      const int a;
 *
 * `a' must NOT be a typedef in the above example. 
 *
 * The presence of `{}' in the following rules  indicates  points 
 * at  which  the symbol table MUST be updated so that the tokenizer 
 * can IMMEDIATELY  continue  to  maintain  the  proper  distinction 
 * between a TYPEDEFname and an IDENTIFIER. 
 */
    
default.declaring.list:  /* Can't  redeclare typedef names */
        declaration.qualifier.list   identifier.declarator
            {
		/*
		 * A declaration.qualifier.list can contain storage classes,
		 * the keywords typedef, volatile and const but no base type.
		 */

		EXECUTE_DEBUG(cerr << "declaration qualifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->duplicate();

		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));
		
		general_declaration(base_type_ptr, $2);
		
		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    }
            initializer.opt
            { 
		$2->initializers = $4;
		$$ = $<variable_declaration_ptr>3;
	    }
        | type.qualifier.list        identifier.declarator
            {
		/*
		 * A type.qualifier.list contains at least one of the keywords
		 * const and volatile but no base type.
		 */

		EXECUTE_DEBUG(cerr << "Type specifier declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->duplicate();
		
		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

		general_declaration(base_type_ptr, $2);
		
		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
          initializer.opt
            {     
		$2->initializers = $4;
		$$ = $<variable_declaration_ptr>3;
	    }
        | default.declaring.list ',' identifier.declarator
            {
		/*
		 * The base type for the identifier being declared here is
		 * entered in the base type field for the declaring list.
		 * Extract it, copy it, process its type elements then use
		 * it to build the type for the identifier being declared 
		 * here.  Add the default base type int to the base type
		 * extracted from the declaring list and use it to build
		 * type correct type.
		 */

		EXECUTE_DEBUG(cerr << "Adding declarator: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " to a declaration!\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->declaration_type->duplicate();
		
		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

		general_declaration(base_type_ptr, $3);
		
		//  Link the declarator onto the current declaration.
		$1->add_declarator($3);

		$$ = $1;
	    }
          initializer.opt
            {
		$3->initializers = $5;
		$$ = $<variable_declaration_ptr>4;
	    }

        | declaration.qualifier.list constructed.identifier.declarator
            {
		/*
		 * A declaration.qualifier.list can contain storage classes,
		 * the keywords typedef, volatile and const but no base type.
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->duplicate();
		
		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

		general_declaration(base_type_ptr, $2);
		
		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | type.qualifier.list        constructed.identifier.declarator
            {
		/*
		 * A type.qualifier.list contains at least one of the keywords
		 * const and volatile but no base type.
		 */

		EXECUTE_DEBUG(cerr << "Type specifier declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->duplicate();
		
		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

		general_declaration(base_type_ptr, $2);
		
		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | default.declaring.list ',' constructed.identifier.declarator
            {
		/*
		 * The base type for the identifier being declared here is
		 * entered in the base type field for the declaring list.
		 * Extract it, copy it, process its type elements then use
		 * it to build the type for the identifier being declared 
		 * here.  Add the default base type int to the base type
		 * extracted from the declaring list and use it to build
		 * type correct type.
		 */

		EXECUTE_DEBUG(cerr << "Adding declarator: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " to a declaration!\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->declaration_type->duplicate();
		
		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

		general_declaration(base_type_ptr, $3);
		
		//  Link the declarator onto the current declaration.
		$1->add_declarator($3);

		$$ = $1;
	    }
        ;

/*
 * Note  how  type.qualifier.list  is  NOT used in the following 
 * productions.   Qualifiers  are   NOT   sufficient   to   redefine 
 * typedef-names (as prescribed by the ANSI C standard).
 */  

declaring.list:
        declaration.specifier  declarator 
            {
		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $2);

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
          initializer.opt
            {     
		$2->initializers = $4;
		$$ = $<variable_declaration_ptr>3;
	    }
        | type.specifier       declarator 
            {
		EXECUTE_DEBUG(cerr << "Type specifier declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $2);

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
          initializer.opt
            {     
		$2->initializers = $4;
		$$ = $<variable_declaration_ptr>3;
	    }
        | basic.type.name      declarator 
	    {
		 /*
		  * The basic.type.name contains a single base type name.
		  */

		EXECUTE_DEBUG(cerr<<"Basic type declaration for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(new type($1->duplicate()), $2);
			
		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(new type($1), $2);
	    } 
	  initializer.opt
            {     
		$2->initializers = $4;
		$$ = $<variable_declaration_ptr>3;
	    }
        | TYPEDEFname          declarator 
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct, union or enum).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $2);

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $2);
	    } 
          initializer.opt
            {
		$2->initializers = $4;
		$$ = $<variable_declaration_ptr>3;
	    }
        | declaring.list ','   declarator 
            {
		/*
		 * The base type for the identifier being declared here is
		 * entered in the base type field for the declaring list.
		 * Abstract it, copy it, process its type elements then use
		 * it to build the type for the identifier being declared 
		 * here.
		 */

		EXECUTE_DEBUG(cerr << "Adding declarator: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " to a declaration!\n";)

		type *base_type_ptr = $1->declaration_type;
		    
		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(base_type_ptr->duplicate(), $3);
		
		//  Link the declarator onto the current declaration.
		$1->add_declarator($3);

		$$ = $1;
	    }
          initializer.opt
            {
		$3->initializers = $5;
		$$ = $<variable_declaration_ptr>4;
	    }
        | declaration.specifier constructed.declarator
            {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $2);
		
		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | type.specifier        constructed.declarator
            {
		/*
		 * A type.specifier contains at least one of the keywords
		 * const and volatile and the base type.
		 */

		EXECUTE_DEBUG(cerr << "Type specifier declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $2);

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | basic.type.name       constructed.declarator
	    {
		/*
		 * The basic.type.name contains a single base type name.  This
		 * must be made into a type and entered in the symbol table
		 * entry for the identifier being declarared.
		 */

		EXECUTE_DEBUG(cerr<<"Basic type declaration for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(new type($1->duplicate()), $2);

		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(new type($1), $2);
	    } 
        | TYPEDEFname           constructed.declarator
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $2);
			
		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $2);
	    } 
        | declaring.list ','    constructed.declarator
            {
		/*
		 * The base type for the identifier being declared here is
		 * entered in the base type field for the declaring list.
		 * Abstract it, copy it, process its type elements then use
		 * it to build the type for the identifier being declared 
		 * here.
		 */

		EXECUTE_DEBUG(cerr << "Adding declarator: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " to a declaration!\n";)

		type *base_type_ptr = $1->declaration_type;

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(base_type_ptr->duplicate(), $3);

		//  Link the declarator onto the current declaration.
		$1->add_declarator($3);

		$$ = $1;
	    }
        ;

/******************************************************************************
 *
 *				Constructed Declarators
 *
 *****************************************************************************/

/*
 * Declarators  with  parenthesized  initializers  present a big 
 * problem.  Typically a declarator that looks  like:  "*a(...)"  is 
 * supposed  to bind FIRST to the "(...)", and then to the "*". This 
 * binding presumes that the "(...)" stuff  is  a  prototype  scope.  
 * With  constructed  declarators,  we  must (officially) finish the 
 * binding to the "*" (finishing forming a good declarator) and THEN 
 * connect with the argument list.  Unfortunately, by  the  time  we 
 * realize  it  is  an  argument  list (and not a prototype) we have 
 * pushed the separate declarator tokens "*a" onto  the  yacc  stack 
 * WITHOUT combining them. The solution is to use odd productions to 
 * carry   the   incomplete  declarator  along  with  the  "argument 
 * expression list" back up the yacc stack.  We would then  actually 
 * instantiate  the  symbol  table after we have fully decorated the 
 * symbol with all the leading "*" stuff. Actually, since  we  don't 
 * have  all  the  type  information in one spot till we reduce to a 
 * declaring.list, this delay is not a problem.  Note that  ordinary 
 * initializers  REQUIRE (ANSI C Standard) that the symbol be placed 
 * into the symbol table BEFORE its initializer is read, but in  the 
 * case  of  parenthesized  initializers,  this  is not possible (we 
 * don't even know we have  an  initializer  till  have  passed  the 
 * opening "(". )
 */

constructed.declarator:
        nonunary.constructed.identifier.declarator
        | constructed.paren.typedef.declarator 
        | simple.paren.typedef.declarator '(' argument.expression.list ')'  
            {
		$1->parenthesized_initializers = $3;
		$$ = $1;
	    }
        | simple.paren.typedef.declarator postfixing.abstract.declarator 
                                          '(' argument.expression.list ')'  
                                          /* semantic error*/
            {
		/*
		 * This construction is a semanctic error.  Log the error and
		 * build the construct anyway so that the parse can continue.
		 */

		ostream error(max_message_size, message_buffer);
		error << "Illegal parenthesized initializer following ";
		error << "function or array declaration!\n";
		error.put('\0');
		yyerror(message_buffer);

		if ($1->declarator_type == NULL)
		    $1->declarator_type = $2;
		else
		    $1->declarator_type->add_type($2);
		$1->parenthesized_initializers = $4;
		$$ = $1;
	    }
        | constructed.parameter.typedef.declarator 
        | indirect.or.reference constructed.declarator
            {    
		$2->add_type(new type(new type_modifier($1)));
		$$ = $2;
	    }
        | pointer.operator      constructed.declarator
            {    
		$2->add_type($1);
		$$ = $2;		
	    }
        ;

constructed.paren.typedef.declarator:
        '(' paren.typedef.declarator ')'                                       
                    '(' argument.expression.list ')'  
            {
		$2->parenthesized_initializers = $5;
		$$ = $2;
	    }
        | '(' paren.typedef.declarator ')' postfixing.abstract.declarator      
                   '(' argument.expression.list ')'  
            {
		if ($2->declarator_type == NULL)
		    $2->declarator_type = $4;
		else
		    $2->declarator_type->add_type($4);
		$2->parenthesized_initializers = $6;
		$$ = $2;
	    }
        | '(' simple.paren.typedef.declarator postfixing.abstract.declarator ')' 
                   '(' argument.expression.list ')'  
            {
		if ($2->declarator_type == NULL)
		    $2->declarator_type = $3;
		else
		    $2->declarator_type->add_type($3);
		$2->parenthesized_initializers = $6;
		$$ = $2;
	    }
        | '(' TYPEDEFname postfixing.abstract.declarator ')'                   
                   '(' argument.expression.list ')'  
            {
		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($2, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($2);
		
		$$ = new declarator(id_ptr, $3);
		$$->parenthesized_initializers = $6;
	    }
        ;

constructed.parameter.typedef.declarator:
        TYPEDEFname    '(' argument.expression.list ')' 
            {
		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($1, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($1);
		
		$$ = new declarator(id_ptr);
		$$->parenthesized_initializers = $3;
	    }
        | TYPEDEFname postfixing.abstract.declarator                       
                       '(' argument.expression.list ')'  /* semantic error */
            {
		/*
		 * This construction is a semanctic error.  Log the error and
		 * build the construct anyway so that the parse can continue.
		 */

		ostream error(max_message_size, message_buffer);
		error << "Illegal parenthesized initializer following ";
		error << "function or array declaration!\n";
		error.put('\0');
		yyerror(message_buffer);

		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($1, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($1);
		
		$$ = new declarator(id_ptr, $2);
		$$->parenthesized_initializers = $4;
	    }
        | '(' clean.typedef.declarator ')'                                 
                       '(' argument.expression.list ')'
            {
		$2->parenthesized_initializers = $5;
		$$ = $2;
	    }
        | '(' clean.typedef.declarator ')'  postfixing.abstract.declarator
                       '(' argument.expression.list ')'
            {
		if ($2->declarator_type == NULL)
		    $2->declarator_type = $4;
		else
		    $2->declarator_type->add_type($4);
		$2->parenthesized_initializers = $6;
		$$ = $2;
	    }
        ;



constructed.identifier.declarator:
        nonunary.constructed.identifier.declarator
        | indirect.or.reference constructed.identifier.declarator
            {    
		$2->add_type(new type(new type_modifier($1)));
		$$ = $2;
	    }
        | pointer.operator      constructed.identifier.declarator
            {    
		$2->add_type($1);
		$$ = $2;		
	    }
        ;

/*
 * The  following  are  restricted to NOT begin with any pointer
 * operators.  This includes both "*" and "T::*"  modifiers.   Aside 
 * from   this   restriction,   the   following   would  have  been: 
 * identifier.declarator '(' argument.expression.list ')' 
 */

nonunary.constructed.identifier.declarator:
        paren.identifier.declarator   '(' argument.expression.list ')'
            {
		$1->parenthesized_initializers = $3;
		$$ = $1;
	    }
        | paren.identifier.declarator postfixing.abstract.declarator   
                       '(' argument.expression.list ')'  /* semantic error*/
            {
		/*
		 * This construction is a semanctic error.  Log the error and
		 * build the construct anyway so that the parse can continue.
		 */

		ostream error(max_message_size, message_buffer);
		error << "Illegal parenthesized initializer following ";
		error << "function or array declaration!\n";
		error.put('\0');
		yyerror(message_buffer);

		if ($1->declarator_type == NULL)
		    $1->declarator_type = $2;
		else
		    $1->declarator_type->add_type($2);
		$1->parenthesized_initializers = $4;
		$$ = $1;
	    }
        | '(' unary.identifier.declarator ')' 
                       '(' argument.expression.list ')'
            {
		$2->parenthesized_initializers = $5;
		$$ = $2;
	    }
        | '(' unary.identifier.declarator ')' postfixing.abstract.declarator  
                       '(' argument.expression.list ')'
            {
		if ($2->declarator_type == NULL)
		    $2->declarator_type = $4;
		else
		    $2->declarator_type->add_type($4);		
		$2->parenthesized_initializers = $6;
		$$ = $2;
	    }
        ;


/******************************************************************************
 *
 *		      	Declaration Specifiers and Qualifiers
 *
 *****************************************************************************/

declaration.specifier:
        basic.declaration.specifier          /* arithmetic or void      */
        | sue.declaration.specifier          /* struct/union/enum/class */
        | typedef.declaration.specifier      /* typedef                 */
        ;

type.specifier:
        basic.type.specifier                 /* arithmetic or void      */
        | sue.type.specifier                 /* struct/Union/Enum/Class */
        | typedef.type.specifier             /* typedef                 */
        ;


declaration.qualifier.list:  /* storage class and optional const/volatile */
        storage.class
            {
		$$ = new type($1);
	    }
        | type.qualifier.list storage.class
            {
		$1->add_type_element($2);
		$$ = $1;
	    }
        | declaration.qualifier.list declaration.qualifier
            {
		$1->add_type_element($2);
		$$ = $1;
	    }
        ;

type.qualifier.list:
        type.qualifier
            {
		$$ = new type($1);
	    }
        | type.qualifier.list type.qualifier
            {
		$1->add_type_element($2);
		$$ = $1;
	    }
        ;

declaration.qualifier:
        type.qualifier                  /* const or volatile */
        | storage.class
	    {
		$$ = $<type_modifier_ptr>1;
	    }
        ;

type.qualifier:
        CONST
            {
		$$ = new type_modifier($1);
	    }
        | VOLATILE
            {
		$$ = new type_modifier($1);
	    }
        ;

basic.declaration.specifier:      /*Storage Class+Arithmetic or void*/
        basic.type.specifier  storage.class
            {
		$$ = new type($2);
		$$->add_type($1);
	    }
        | basic.type.name     storage.class
            {
		$$ = new type($2);
		$$->add_type_element($1);
	    }
        | declaration.qualifier.list basic.type.name
            {
		$$ = $1;
		$$->add_type_element($2);
	    }
        | basic.declaration.specifier declaration.qualifier
            {
		$$ = new type($2);
		$$->add_type($1);
	    }
        | basic.declaration.specifier basic.type.name
            {
		$$ = $1;
		$$->add_type_element($2);
	    }
        ;

basic.type.specifier:
        basic.type.name        basic.type.name /* Arithmetic or void */
	    {
		$$ = new type($1);
		$$->add_type_element($2);
	    }
        | basic.type.name      type.qualifier
	    {
		$$ = new type($2);
		$$->add_type_element($1);
	    }
        | type.qualifier.list  basic.type.name
	    {
		$1->add_type_element($2);
		$$ = $1;
	    }
        | basic.type.specifier type.qualifier
	    {
		$$ = new type($2);
		$$->add_type($1);
	    }
        | basic.type.specifier basic.type.name
	    {
		$$ = $1;
		$$->add_type_element($2);
	    }
        ;

sue.declaration.specifier: /* Storage Class + struct/union/enum/class */
        sue.type.specifier storage.class
	    {
		$1->add_type_element($2);
		$$ = $1;
	    }
        | declaration.qualifier.list elaborated.type.name
	    {
		$1->add_type($2);
		$$ = $1;
	    }
        | sue.declaration.specifier declaration.qualifier
	    {
		$1->add_type_element($2);
		$$ = $1;
	    }
        ;

sue.type.specifier:
        elaborated.type.name              /* struct/union/enum/class */
        | type.qualifier.list elaborated.type.name
	    {
		$1->add_type($2);
		$$ = $1;
	    }
        | sue.type.specifier type.qualifier
	    {
		$1->add_type_element($2);
		$$ = $1;
	    }
        ;

typedef.declaration.specifier:       /*Storage Class + typedef types */
        typedef.type.specifier storage.class
        | TYPEDEFname          storage.class
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($1);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		
		type *type_ptr = new type($2);
		type_ptr->add_type(typedef_ptr);
		$$ = type_ptr;
	    }
        | declaration.qualifier.list TYPEDEFname
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		
		$1->add_type(typedef_ptr);
		$$ = $1;
	    }
        | typedef.declaration.specifier declaration.qualifier
            {
		type *temp_type = new type($2);
		temp_type->add_type($1);
		$$ = temp_type;
	    }
        ;

typedef.type.specifier:              /* typedef types */
        TYPEDEFname              type.qualifier
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($1);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		
		type *type_ptr = new type($2);
		type_ptr->add_type(typedef_ptr);
		$$ = type_ptr;
	    }
        | type.qualifier.list    TYPEDEFname
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		
		$1->add_type(typedef_ptr);
		$$ = $1;
	    }
        | typedef.type.specifier type.qualifier
            {
		type *temp_type = new type($2);
		temp_type->add_type($1);
		$$ = temp_type;
	    }
        ;

/******************************************************************************
 *
 *				Storage Classes
 *
 *****************************************************************************/

/* 
 * There are really several  distinct  sets  of  storage.classes.  
 * The  sets  vary  depending  on whether the declaration is at file 
 * scope, is a  declaration  within  a  struct/class,  is  within  a 
 * function body, or in a function declaration/definition (prototype 
 * parameter  declarations).  They  are grouped here to simplify the 
 * grammar,  and  can  be  semantically  checked.   Note  that  this 
 * approach  tends to ease the syntactic restrictions in the grammar 
 * slightly, but allows for future language development,  and  tends 
 * to  provide  superior  diagnostics  and  error  recovery (i.e.: a 
 * syntax error does not disrupt the parse).
 *
 *             File    File    Member  Member  Local   Local  Formal
 *             Var     Funct   Var     Funct   Var     Funct  Params
 * TYPEDEF         x       x       x       x       x       x
 * EXTERN          x       x                       x       x
 * STATIC          x       x       x       x       x
 * AUTO                                            x              x
 * REGISTER                                        x              x
 * FRIEND                                  x       
 * INLINE                  x               x               x
 * VIRTUAL                                 x               x
 */

storage.class:
        TYPEDEF 
            {
		$$ = new storage_class($1);
	    }
        | EXTERN
            {
		$$ = new storage_class($1);
	    }
        | STATIC 
            {
		$$ = new storage_class($1);
	    }
        | AUTO
            {
		$$ = new storage_class($1);
	    }
        | REGISTER
            {
		$$ = new storage_class($1);
	    }
        | FRIEND
            {
		$$ = new storage_class($1);
	    }
        | INLINE  
            {
		$$ = new storage_class($1);
	    }
        | VIRTUAL 
            {
		$$ = new storage_class($1);
	    }
        ;

/******************************************************************************
 *
 *				Base Types
 *
 *****************************************************************************/

basic.type.name:
        VOID
            {
		$$ = new base_type($1);
	    }
        | CHAR
            {
		$$ = new base_type($1);
	    }
        | SHORT
            {
		$$ = new base_type($1);
	    }
        | INT
            {
		$$ = new base_type($1);
	    }
        | LONG
            {
		$$ = new base_type($1);
	    }
        | FLOAT
            {
		$$ = new base_type($1);
	    }
        | DOUBLE
            {
		$$ = new base_type($1);
	    }
        | SIGNED
            {
		$$ = new base_type($1);
	    }
        | UNSIGNED
            {
		$$ = new base_type($1);
	    }
        ;

/******************************************************************************
 *
 *		 	Class, Struct, Union and Enum Declarations
 *
 *****************************************************************************/

elaborated.type.name:
        aggregate.name
	    {
		identifier *id_ptr = $1->identifier_ptr;
		type *type_ptr = id_ptr->type_ptr->duplicate();
		type_ptr->expand_typedefs();
		if (!$1->print_body)
		    ((base_type *)type_ptr->get_type_element())->print_body
			= FALSE;
		$$ = type_ptr;
	    }
        | enum.name
	    {
		identifier *id_ptr = $1->enum_id_ptr;
		type *type_ptr = id_ptr->type_ptr->duplicate();
		type_ptr->expand_typedefs();
		if (!$1->print_body)
		    ((base_type *)type_ptr->get_type_element())->print_body
			= FALSE;
		$$ = type_ptr;
	    }
        ;


/* 
 * Since the expression "new type.name" MIGHT use  an  elaborated 
 * type  and a derivation, it MIGHT have a ':'.  This fact conflicts 
 * with the requirement that a new expression can be placed  between 
 * a '?' and a ':' in a conditional expression (at least it confuses 
 * most LR(1) parsers). 
 *
 *  The derivation.opt reduction was placed DELIBERATELY in front 
 * of aggregate.name to resolve  a  reduce-reduce  on  '{'  conflict 
 * properly 
 */
    
derivation.opt:
        /* nothing */
            {
		$$ = NULL;
	    }
        | ':' derivation.list
            {
		$$ = $2;
	    }
        ;

/*
 * The  intermediate  actions  {}  represent points at which the 
 * database of typedef names  must  be  updated  in  C++.   This  is 
 * critical to the lexer, which must begin to tokenize based on this 
 * new information. 
 */

aggregate.name:
        aggregate.key derivation.opt 
            {
		/*
		 * An unnamed class/struct/union definition is being parsed.  
		 *
		 * aggregate.key - must be one of the following keywords:
		 *                 CLASS, STRUCT, UNION, MENTAT, PERSISTENT or
		 *		   REGULAR.
		 *
		 * derivation.opt - the optional class derivation list.
		 *
		 * member.declaration.list.opt - optional statement list
		 *                               containing the member 
		 *                               declarations for this class/
		 *                               struct/union.
		 */

		/*
		 * The csu_token indicates whether this is a class, struct or 
		 * union. Aggregate.key will be one of these or one of the 
		 * mentat key words.  The various mentat class types have
		 * csu_token CLASS.  Check to see if aggregate.key is one of 
		 * the mentat key words, if so set the csu_token to be CLASS.  
		 * If not the csu_token is just the aggregate.key.
		 */

		TOKEN csu_token = $1;

		if ($1 == MENTAT || $1 == PERSISTENT || $1 == REGULAR)
		{
		    /*
		     * This is an error.  An unnamed struct or class cannot be 
		     * a mentat class.
		     */

		    cerr << program_name << ": line " << linenum;
		    cerr <<" Unnamed class/struct/union declared as mentat!\n";
		    csu_token = CLASS;
		}

		/*
		 * Set the global variable current_access_token to be PUBLIC or
		 * PRIVATE depending on whether this is a struct or a class.
		 * This global variable controls whether declarations are by 
		 * default marked as PUBLIC, PRIVATE or PROTECTED.  The default
		 * for classes is PRIVATE, for structs and unions it is PUBLIC.
		 */

		if (csu_token == STRUCT || csu_token == UNION)
		    current_access_token = PUBLIC;
		else
		    current_access_token = PRIVATE;


		/*
		 * Create a class definition object for the class/struct/union.
		 * Use an unnamed identifier object in the class definition.
		 */

		identifier *id_ptr = new identifier(NULL, csu_token);
		
		class_definition *csu_definition_ptr = 
		    new class_definition(csu_token,id_ptr,new symbol_table,$2);

		/*
		 * Set up the scope.  The member function set_scope pushes the
		 * appropriate symbol tables for the base classes onto the 
		 * symbol table stack.  Thus all inerited members will be in
		 * scope.  It also pushes the symbol table for this class so
		 * that the members will be entered in the correct symbol 
		 * table.
		 */

		csu_definition_ptr->set_scope();

		// Set the type of the class/struct/union identifier.
		id_ptr->type_ptr = new type(new base_type(csu_token, id_ptr));

		// Set the csu definition field in the identifier object.
		id_ptr->csue_definition_ptr = csu_definition_ptr;

		// Set the current class pointer so THIS is typed correctly.
		current_class.push(csu_definition_ptr);
		
		// Return the class/struct/union definition for use below.
		$<class_definition_ptr>$ = csu_definition_ptr;
	    }
                '{' member.declaration.list.opt '}'

	    {
		/*
		 * The class/struct/union definition body has been parsed.  
		 * Add it to the class/struct/union definition created above.
		 * Then reset the scope now that the members have been 
		 * declared.
		 */

		$<class_definition_ptr>3->declarations = $5;

		// Reset the scope.
		TABLE_DEBUG(scope_stack.dump_top();)
		$<class_definition_ptr>3->reset_scope();
		
		// Pop the current class so that THIS is typed correctly.
		current_class.pop();
		
		// Return the class/struct/union definition.
		$$ = $<class_definition_ptr>3;
	    }
        | aggregate.key identifier.or.typedef.name derivation.opt 
            {
		/*
		 * A class/struct/union definition is being parsed.  
		 *
		 * aggregate.key - must be one of the following keywords:
		 *                 CLASS, STRUCT, UNION, MENTAT, PERSISTENT or
		 *		   REGULAR.
		 *
		 * identifier.or.typedef.name - string representation of the
		 *                              class/struct/union name.
		 * 
		 *
		 * derivation.opt - the optional class derivation list.
		 *
		 * member.declaration.list.opt - optional statement list
		 *                               containing the member 
		 *                               declarations for this class/
		 *                               struct/union.
		 */

		/*
		 * The csu_token indicates whether this is a class, struct or 
		 * union. Aggregate.key will be one of these or one of the 
		 * mentat key words.  The various mentat class types have 
		 *  nu_token CLASS.  Check to see if aggregate.key is one of 
		 * the mentat key words, if so set the csu_token to be CLASS.  
		 * If not the csu_token is just the aggregate.key.  If this is 
		 * a mentat class set the mentat_flag to true.
		 */

		TOKEN csu_token     = $1;
		boolean mentat_flag = FALSE;

		if ($1 == MENTAT || $1 == PERSISTENT || $1 == REGULAR)
		{
		    csu_token   = CLASS;
		    mentat_flag = TRUE;
		}

		/*
		 * Set the global variable current_access_token to be PUBLIC or
		 * PRIVATE depending on whether this is a struct or a class.
		 * This global variable controls whether declarations are by 
		 * default marked as PUBLIC, PRIVATE or PROTECTED.  The default
		 * for classes is PRIVATE, for structs and unions it is PUBLIC.
		 */

		if (csu_token == STRUCT || csu_token == UNION)
		    current_access_token = PUBLIC;
		else
		    current_access_token = PRIVATE;

		/*
		 * Enter the csu name in the symbol table.  There can be two
		 * reasons for this name already being in the symbol table.  
		 * It will be in the symbol table if the csu was declared 
		 * before it was defined, in which case there will be no csu
		 * definition associated with the csu name.  It will also be
		 * in the symbol table if the identifier was used for something
		 * else, which would be an error. Figure out which case this is
		 * and take the appropriate action.
		 */

		identifier *id_ptr = scope_stack.insert($2, csu_token);
		if (id_ptr == NULL)
		{
		    // Lookup the existing identifier.
		    id_ptr = scope_stack.lookup($2);

		    /*
		     * If the TYPEDEFname does not match this csu definition 
		     * or if the csu has already been defined, log the error
		     * and ignore the previous declaration or definition.
		     */

		    if (match_token(id_ptr->id_token, csu_token))
		    {
			/*
			 * The TYPEDEFname matches this csu.  If this is a  
			 * forward declaration there is no error.  If the csu
			 * is being redefined, log the error and ignore the
			 * previous definition.
			 */

			if (id_ptr->csue_definition_ptr != NULL)
			{
			    cerr << program_name << ": line " << linenum;
			    cerr << " Class, struct or union: " << $2;
			    cerr << " being redefined!  Ignoring previous ";
			    cerr << "definition!\n";
			    id_ptr->csue_definition_ptr = NULL;
			}
		    }
		    else
		    {
			/*
			 * The csu identifier is being redefined, log the error
			 * and ignore the previous definition.
			 */

			cerr << program_name << ": line " << linenum;
			cerr << " Identifier: " << $2;
			cerr << " being redefined! Ignoring previous ";
			cerr << "definition or declaration!\n";
			id_ptr = scope_stack.remove(id_ptr);
			id_ptr = scope_stack.insert($2, csu_token);
		    }
		}

		//Create a class definition object for this class/struct/union.
		class_definition *csu_definition_ptr = 
		    new class_definition(csu_token,id_ptr,new symbol_table,$3);

		/*
		 * Set up the scope.  The member function set_scope pushes the
		 * appropriate symbol tables for the base classes onto the 
		 * symbol table stack.  Thus all inerited members will be in
		 * scope.  It also pushes the symbol table for this csu so
		 * that the members will be entered in the correct symbol 
		 * table.
		 */

		csu_definition_ptr->set_scope();

		// Set the type of the csu identifier.

		// Create the new base_type object.
		base_type *base_type_ptr = 
		    new base_type(TYPEDEFname, id_ptr);
		base_type_ptr->typedef_type_ptr = 
		    new type(new base_type(csu_token, id_ptr));
		id_ptr->type_ptr = new type(base_type_ptr);

		/*
		 * If this is a mentat class, set the mentat flag and enter its
		 * definition on the mentat classes list.
		 */

		if (mentat_flag == TRUE) 
		{
		    csu_definition_ptr->mentat_class_flag = TRUE;
		    if ($1 == PERSISTENT)
			csu_definition_ptr->persistent_flag = TRUE;
		}			

		// Set the csu definition field in the identifier object.
		id_ptr->csue_definition_ptr = csu_definition_ptr;
		id_ptr->redeclare_name_flag = FALSE;

		// Set the current class pointer so THIS is typed correctly.
		current_class.push(csu_definition_ptr);
		
		// Return the class/struct/union definition for use below.
		$<class_definition_ptr>$ = csu_definition_ptr;
	    }

                '{' member.declaration.list.opt '}'

            {
		/*
		 * The class/struct/union definition body has been parsed.  
		 * Add it to the class/struct/union definition created above.
		 * Then reset the scope now that the members have been 
		 * declared.
		 */

		$<class_definition_ptr>4->declarations = $6;

		// Reset the scope.
		TABLE_DEBUG(scope_stack.dump_top();)
		$<class_definition_ptr>4->reset_scope();
		
		// Pop the current class so that THIS is typed correctly.
		current_class.pop();
		
		// Return the class/struct/union definition.
		$$ = $<class_definition_ptr>4;
	    }
        | aggregate.key identifier.or.typedef.name
            {
		/*
		 * A class/struct/union definition is being parsed.  
		 *
		 * aggregate.key - must be one of the following keywords:
		 *                 CLASS, STRUCT, UNION, MENTAT, PERSISTENT or
		 *		   REGULAR.
		 *
		 * identifier.or.typedef.name - string representation of the
		 *                              class/struct/union name.
		 */

		/*
		 * The csu_token indicates whether this is a class, struct or 
		 * union. Aggregate.key will be one of these or one of the 
		 * mentat key words.  The various mentat class types have 
		 * csu_token CLASS.  Check to see if aggregate.key is one of 
		 * the mentat key words, if so set the csu_token to be CLASS.
		 * If not the csu_token is just the aggregate.key.  If this is
		 * a mentat class set the mentat_flag to true.
		 */

		TOKEN csu_token     = $1;
		boolean mentat_flag = FALSE;

		if ($1 == MENTAT || $1 == PERSISTENT || $1 == REGULAR)
		{
		    csu_token   = CLASS;
		    mentat_flag = TRUE;
		}

		/*
		 * Lookup the type name in the symbol table.  If it is not
		 * there then this must be a forward class declaration so
		 * enter the name in the symbol table.  If it is already
		 * in the symbol table, it may be a declaration of a new class
		 * with a type name from an enclosing scope, for example:
		 *
		 * struct foo {	 ...  };   // Global foo	
		 * void bar(){
		 *     struct foo;         // Local foo
		 * }
		 *
		 * Or it may be the anachronistic 'struct typename' way of 
		 * specifying a type name, for example:
		 *
		 * struct foo {	 ...  };   // Global foo	
		 * void bar(){
		 *     struct foo *object; // Reference to global foo
		 * }
		 *
		 * It cannot be determined at this point in the parse, which
		 * form it is so assume it is the 'struct typename' form and if
		 * the next character after the 'struct typename' is a ';' then
		 * make the appropriate entry into the symbol table.  The rules
		 * that determine this are:
		 *
		 * declaration:
		 *    sue.declaration.specifier ';'
		 *    | sue.type.specifier ';'
		 *
		 * and
		 *
		 * member.declaration:
	         *     ...
                 *     | sue.type.specifier ';'
		 *
		 * To do the appropriate entries identifier objects have the 
		 * following member function and variables:
	         * 
		 * redeclare_name_flag - This variable indicates whether or not
		 *                       this type name should be redeclared if
		 *                       followed by a ';'.
		 *
		 * redeclare_name()    - This function enters the type name
		 *                       that is being redeclared into the 
		 *                       symbol table.
		 *
		 * redeclare_csu_token - This variable contains the csu_token
		 *                       used when redeclaring the type name.
		 *
		 * Redeclare_name_flag and redeclare_csu_token must be set here
		 * to appropriate values.
		 */

		identifier *id_ptr = scope_stack.lookup_type($2);
		if (id_ptr == NULL)
		{
		    // The type name is not in the symbol table.

		    // Insert the type name identifier into symbol table.
		    id_ptr = scope_stack.insert($2, csu_token);

		    // Create the new base_type object.
		    base_type *base_type_ptr = 
			new base_type(TYPEDEFname, id_ptr);
		    base_type_ptr->typedef_type_ptr = 
			new type(new base_type(csu_token, id_ptr));
		    id_ptr->type_ptr = new type(base_type_ptr);
		    id_ptr->redeclare_name_flag = FALSE;
		}
		else
		{
		    // A type with this name is already in the symbol table.

		    id_ptr->redeclare_name_flag = TRUE;
		    id_ptr->redeclare_csu_token = csu_token;
		}
		
		//Create a class definition object.
		class_definition *csu_definition_ptr = 
		    new class_definition(csu_token, id_ptr);
		    
		// There is no body with this csu declaration.
		csu_definition_ptr->print_body = FALSE;

		// Return the class/struct/union definition.
		$$ = csu_definition_ptr;
	    }
        ;

derivation.list:
        parent.class
	    {	
		$$ = $1;
	    }
        | derivation.list ',' parent.class
	    {	
		$1->add_derivation($3);
	    }
        ;

parent.class:
        TYPEDEFname
	    {
		/*
		 * Lookup the TYPEDEFname, which is a class/struct/union name, 
		 * in the symbol table.  It must be in the symbol table for the
		 * lexer to have  identified this as a typedef name, so no need
		 * to check the return value.  However make sure this is indeed
		 * a class/struct/union.  Get the class/struct/union name's
		 * definition and use it to build a derivation list.
		 */

		identifier *csu_id_ptr = scope_stack.lookup($1);
		
		// Make sure this is a class/struct/union.
		TOKEN csu_token = csu_id_ptr->id_token;
		if (csu_token != CLASS && csu_token != STRUCT && 
		    csu_token != UNION)
		{
		    /*
		     * A non-class/struct/union identifier is being used in a
		     * derivation list.  Log an error and since I don't know
		     * how to recover from this simply exit.
		     */
		    cerr << program_name << ": line " << linenum;
		    cerr << " Non-class/struct/union name : ";
		    cerr << csu_id_ptr->id_string << "in derivation list.";
		    cerr << "Mplfront will abort!\n";
		    exit(-1);
		}
		
		    
		class_definition *class_definition_ptr = 
		    csu_id_ptr->csue_definition_ptr;
		if (class_definition_ptr == NULL)
		{
		    /*
		     * Log an error, a forward defined class is being used in a 
		     * derivation list.
		     */
		    cerr << program_name << ": line " << linenum;
		    cerr << " Undefined class/struct/union : ";
		    cerr << csu_id_ptr->id_string << "in derivation list.";
		    cerr << "Mplfront will abort!\n";
		    exit(-1);
		}
		$$ = new derivation_list(class_definition_ptr);
	    }
        | VIRTUAL access.specifier.opt TYPEDEFname
	    {	
		/*
		 * Lookup the TYPEDEFname, which is a class/struct/union name, 
		 * in the symbol table.  It must be in the symbol table for the
		 * lexer to have  identified this as a typedef name, so no need
		 * to check the return value.  However make sure this is indeed
		 * a class/struct/union.  Get the class/struct/union name's
		 * definition and use it to build a derivation list.
		 */

		identifier *csu_id_ptr = scope_stack.lookup($3);
		
		// Make sure this is a class/struct/union.
		TOKEN csu_token = csu_id_ptr->id_token;
		if (csu_token != CLASS && csu_token != STRUCT && 
		    csu_token != UNION)
		{
		    /*
		     * A non-class/struct/union identifier is being used in a
		     * derivation list.  Log an error and since I don't know
		     * how to recover from this simply exit.
		     */
		    cerr << program_name << ": line " << linenum;
		    cerr << " Non-class/struct/union name : ";
		    cerr << csu_id_ptr->id_string << "in derivation list.";
		    cerr << "Mplfront will abort!\n";
		    exit(-1);
		}
		
		    
		class_definition *class_definition_ptr = 
		    csu_id_ptr->csue_definition_ptr;
		if (class_definition_ptr == NULL)
		{
		    /*
		     * Log an error, a forward defined class is being used in a 
		     * derivation list.
		     */
		    cerr << program_name << ": line " << linenum;
		    cerr << " Undefined class/struct/union : ";
		    cerr << csu_id_ptr->id_string << "in derivation list.";
		    cerr << "Mplfront will abort!\n";
		    exit(-1);
		}
		$$ = new derivation_list(class_definition_ptr);
		$$->virtual_flag = TRUE;
		$$->access_specifier = $2;
	    }
        | access.specifier virtual.opt TYPEDEFname
	    {	
		/*
		 * Lookup the TYPEDEFname, which is a class/struct/union name, 
		 * in the symbol table.  It must be in the symbol table for the
		 * lexer to have  identified this as a typedef name, so no need
		 * to check the return value.  However make sure this is indeed
		 * a class/struct/union.  Get the class/struct/union name's
		 * definition and use it to build a derivation list.
		 */

		identifier *csu_id_ptr = scope_stack.lookup($3);
		
		// Make sure this is a class/struct/union.
		TOKEN csu_token = csu_id_ptr->id_token;
		if (csu_token != CLASS && csu_token != STRUCT && 	
		    csu_token != UNION)
		{
		    /*
		     * A non-class/struct/union identifier is being used in a
		     * derivation list.  Log an error and since I don't know
		     * how to recover from this simply exit.
		     */
		    cerr << program_name << ": line " << linenum;
		    cerr << " Non-class/struct/union name : ";
		    cerr << csu_id_ptr->id_string << "in derivation list.";
		    cerr << "Mplfront will abort!\n";
		    exit(-1);
		}
		
		    
		class_definition *class_definition_ptr = 
		    csu_id_ptr->csue_definition_ptr;
		if (class_definition_ptr == NULL)
		{
		    /*
		     * Log an error, a forward defined class is being used in a 
		     * derivation list.
		     */
		    cerr << program_name << ": line " << linenum;
		    cerr << " Undefined class/struct/union : ";
		    cerr << csu_id_ptr->id_string << "in derivation list.";
		    cerr << "Mplfront will abort!\n";
		    exit(-1);
		}
		$$ = new derivation_list(class_definition_ptr);
		if ($2 != NONE)
		    $$->virtual_flag = TRUE;
		$$->access_specifier = $1;
	    }
        ;

virtual.opt:
        /* nothing */
            {
		$$ = NONE;
	    }
        | VIRTUAL
        ;

access.specifier.opt:
        /* nothing */
            {
		$$ = NONE;
	    }
        | access.specifier
        ;

access.specifier:
        PUBLIC
        | PRIVATE
        | PROTECTED 
        ;

aggregate.key:
        STRUCT
        | UNION
        | CLASS
        | MENTAT CLASS
        | PERSISTENT MENTAT CLASS
        | REGULAR MENTAT CLASS
        ;

/******************************************************************************
 *
 *			Member Declarations and Definitions
 *
 *****************************************************************************/

/*
 * Note that an empty list is ONLY allowed under C++. The grammar 
 * can  be modified so that this stands out.  The trick is to define 
 * member.declaration.list, and have that referenced for non-trivial 
 * lists. 
 */

member.declaration.list.opt:
        /* nothing */ 
            {
		$$ = NULL;
	    }
        | member.declaration.list.opt member.declaration
            {
		if ($1 == NULL)
		    $$ = new statement_list($2);
		else
		{
		    $1->add_statement($2);
		    $$ = $1;
		}
	    }
        ;

member.declaration:
        member.declaring.list ';'
            {
		$$ = $1;
	    }
        | member.default.declaring.list ';'
            {
		$$ = $1;
	    }
        | PUBLIC ':'
             {
		 current_access_token = PUBLIC;
		 $$ = new declaration_access_indicator(PUBLIC);
	     }
        | PRIVATE ':'                       
             {
		 current_access_token = PRIVATE;
		 $$ = new declaration_access_indicator(PRIVATE);
	     }
        | PROTECTED ':'                     
             {
		 current_access_token = PROTECTED;
		 $$ = new declaration_access_indicator(PROTECTED);
	     }
        | new.function.definition member.function.definition.semi.opt
             {
		 EXECUTE_DEBUG(cerr << "New function definition";)
		 EXECUTE_DEBUG(cerr << " in class declaration.\n";)
		 
		 $$ = new member_declaration($1);
	     }
        | constructor.function.in.class
             {
		 EXECUTE_DEBUG(cerr << "Constructor definition";)
		 EXECUTE_DEBUG(cerr << " in class declaration.\n";)
		 
		 $$ = new member_declaration($1);
	     }
        | destructor.function
             {
		 EXECUTE_DEBUG(cerr << "Destructor definition";)
		 EXECUTE_DEBUG(cerr << " in class declaration.\n";)
		 
		 $$ = new member_declaration($1);
	     }
        | sue.type.specifier ';'            /* nested tag elaborations */
            {
		/*
		 * This is the first place the parser can distinguish between
		 * the following situations:
		 *
		 * 1) struct struct_name *variable_name;
		 *
		 * 2) struct struct_name;  // Where struct_name has been
		 *			   // defined previously.
		 *
		 * Case 2 declares a type name, while case 1 requires a lookup.
		 * The variable redeclare_name_flag indicates that this is
		 * case 2.  The member function redeclare_name declares the
		 * name.
		 */

		identifier *id_ptr = $1->get_class_identifier();
		if (id_ptr != NULL && id_ptr->redeclare_name_flag)
		    id_ptr->redeclare_name();
	    		    
		$$ = new csue_declaration($1);
	    }
        | class.rescoped.identifier ';'     /* access modification */ 
	    {
		$$ = new member_declaration($1);
	    }
        | typedef.declaration.specifier ';' /* friend T */
            {
		$$ = new variable_declaration($1);
	    }
        | sue.declaration.specifier ';'     /* friend class C */
            {
		$$ = new csue_declaration($1);
	    }
        ;

/* 
 * An optional semicolon is allowed after an in class member function 
 * definition, constructor or destructor. See ARM p169. - ecl.
 */

member.function.definition.semi.opt:
        /* nothing */
            {
		$$ = NONE;
	    }
        | ';'
            {      
		$$ = ';';
	    }
		
member.default.declaring.list:        /* doesn't redeclare typedef */
        type.qualifier.list          
                identifier.declarator member.pure.opt
            {
		/*
		 * A type.qualifier.list contains at least one of the keywords
		 * const and volatile but no base type.
		 */

		EXECUTE_DEBUG(cerr << "Type specifier declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->duplicate();
		
		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

		general_declaration(base_type_ptr, $2, current_access_token);
		
		if ($3 != NULL)
		    $2->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | declaration.qualifier.list 
                identifier.declarator member.pure.opt /* C++, not ANSI C */
            {
		/*
		 * A declaration.qualifier.list can contain storage classes,
		 * the keywords typedef, volatile and const but no base type.
		 */

		EXECUTE_DEBUG(cerr << "declaration qualifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->duplicate();

		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));
		
		general_declaration(base_type_ptr, $2, current_access_token);
		
		if ($3 != NULL)
		    $2->set_pure_virtual();

		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    }
        | member.default.declaring.list ',' 
                identifier.declarator member.pure.opt
            {
		/*
		 * The base type for the identifier being declared here is
		 * entered in the base type field for the declaring list.
		 * Extract it, copy it, process its type elements then use
		 * it to build the type for the identifier being declared 
		 * here.  Add the default base type int to the base type
		 * extracted from the declaring list and use it to build
		 * type correct type.
		 */

		EXECUTE_DEBUG(cerr << "Adding declarator: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " to a declaration";)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->declaration_type->duplicate();
		
		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

		general_declaration(base_type_ptr, $3, current_access_token);
		
		if ($4 != NULL)
		    $3->set_pure_virtual();

		//  Link the declarator onto the current declaration.
		$1->add_declarator($3);

		$$ = $1;
	    }

        | type.qualifier.list           bit.field.identifier.declarator 
            {
		/*
		 * A type.qualifier.list contains at least one of the keywords
		 * const and volatile but no base type.
		 */

		EXECUTE_DEBUG(cerr << "Type specifier bit field declaration";)
		EXECUTE_DEBUG(cerr << " for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->duplicate();
		
		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

		if ($2->declarator_id != NULL)
		    general_declaration(base_type_ptr,$2,current_access_token);
		
		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | declaration.qualifier.list    bit.field.identifier.declarator 
            {
		/*
		 * A declaration.qualifier.list can contain storage classes,
		 * the keywords typedef, volatile and const but no base type.
		 */

		EXECUTE_DEBUG(cerr << "declaration qualifier bit field");
		EXECUTE_DEBUG(cerr << "declaration for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->duplicate();

		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));
		
		if ($2->declarator_id != NULL)
		    general_declaration(base_type_ptr,$2,current_access_token);
		
		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    }
        | member.default.declaring.list ','  bit.field.identifier.declarator 
            {
		/*
		 * The base type for the identifier being declared here is
		 * entered in the base type field for the declaring list.
		 * Extract it, copy it, process its type elements then use
		 * it to build the type for the identifier being declared 
		 * here.  Add the default base type int to the base type
		 * extracted from the declaring list and use it to build
		 * type correct type.
		 */

		EXECUTE_DEBUG(cerr << "Adding bit field declarator: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " to a declaration";)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Duplicate the type.  This is done because the canonical
		 * type that is associated with the identifier may have a 
		 * different form than the type to be printed with the 
		 * declarator record.
		 */

		type *base_type_ptr = $1->declaration_type->duplicate();
		
		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

                if ($3->declarator_id != NULL)
		    general_declaration(base_type_ptr,$3,current_access_token);
		
		//  Link the declarator onto the current declaration.
		$1->add_declarator($3);

		$$ = $1;
	    }
        ;

member.declaring.list:        /* Can possibly redeclare typedefs */
        type.specifier     declarator member.pure.opt
            {
		EXECUTE_DEBUG(cerr << "Type specifier declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $2);

		if ($3 != NULL)
		    $2->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | basic.type.name  declarator member.pure.opt
	    {
		 /*
		  * The basic.type.name contains a single base type name.
		  */

		EXECUTE_DEBUG(cerr<<"Basic type declaration for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(new type($1->duplicate()), $2);
			
		if ($3 != NULL)
		    $2->set_pure_virtual();

		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(new type($1), $2);
	    } 
        | member.conflict.declaring.item

        | member.declaring.list ',' declarator member.pure.opt
            {
		/*
		 * The base type for the identifier being declared here is
		 * entered in the base type field for the declaring list.
		 * Abstract it, copy it, process its type elements then use
		 * it to build the type for the identifier being declared 
		 * here.
		 */

		EXECUTE_DEBUG(cerr << "Adding declarator: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " to a declaration";)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		type *base_type_ptr = $1->declaration_type;
		type_element *type_element_ptr =
		    base_type_ptr->get_type_element();

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(base_type_ptr->duplicate(), $3);
		
		if ($4 != NULL)
		    $3->set_pure_virtual();

		//  Link the declarator onto the current declaration.
		$1->add_declarator($3);

		$$ = $1;
	    }
        | type.specifier        bit.field.declarator 
            {
		EXECUTE_DEBUG(cerr << "Type specifier bit field declaration";)
		EXECUTE_DEBUG(cerr << " for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

	        if ($2->declarator_id != NULL)
		    general_declaration($1->duplicate(), $2,
					current_access_token);

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | basic.type.name       bit.field.declarator 
	    {
		 /*
		  * The basic.type.name contains a single base type name.
		  */

		EXECUTE_DEBUG(cerr << "Basic type declaration for bit field";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		    if ($2->declarator_id != NULL)
			general_declaration(new type($1->duplicate()), $2,
					    current_access_token);
			
		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(new type($1), $2);
	    } 
        | TYPEDEFname           bit.field.declarator 
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type bit field declaration";)
		EXECUTE_DEBUG(cerr << " for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		if ($2->declarator_id != NULL)
		    general_declaration(typedef_type_ptr->duplicate(), $2,
					current_access_token);
			
		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $2);
	    } 
        | declaration.specifier bit.field.declarator /* C++? not ANSI C */
            {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier bit field ";)
		EXECUTE_DEBUG(cerr << "declaration for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		if ($2->declarator_id != NULL)
		    general_declaration($1->duplicate(), $2, 
					current_access_token);
		
		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | member.declaring.list ',' bit.field.declarator 
            {
		/*
		 * The base type for the identifier being declared here is
		 * entered in the base type field for the declaring list.
		 * Abstract it, copy it, process its type elements then use
		 * it to build the type for the identifier being declared 
		 * here.
		 */

		EXECUTE_DEBUG(cerr << "Adding bit field declarator: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " to a declaration";)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		type *base_type_ptr = $1->declaration_type;
		type_element *type_element_ptr =
		    base_type_ptr->get_type_element();

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		if ($3->declarator_id != NULL)
		    general_declaration(base_type_ptr->duplicate(), $3,
					current_access_token);
		
		//  Link the declarator onto the current declaration.
		$1->add_declarator($3);

		$$ = $1;
	    }
        ;

/*
 * The following conflict with constructors-
 * member.conflict.declaring.item:
 *     TYPEDEFname           declarator member.pure.opt
 *     | declaration.specifier declarator member.pure.opt
 *     ;
 * so we inline expand declarator to get the following productions...
 */

member.conflict.declaring.item:
        TYPEDEFname             identifier.declarator           member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration";)
		EXECUTE_DEBUG(cerr << " for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $2,
				    current_access_token);

		if ($3 != NULL)
		    $2->set_pure_virtual();
			
		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $2);
	    } 
        | TYPEDEFname           parameter.typedef.declarator    member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration";)
		EXECUTE_DEBUG(cerr << " for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $2,
				    current_access_token);

		if ($3 != NULL)
		    $2->set_pure_virtual();
			
		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $2);
	    } 
        | declaration.specifier identifier.declarator           member.pure.opt
            {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration qualifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $2);
		
		if ($3 != NULL)
		    $2->set_pure_virtual();

		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | declaration.specifier parameter.typedef.declarator    member.pure.opt
            {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration qualifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $2);
		
		if ($3 != NULL)
		    $2->set_pure_virtual();

		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | TYPEDEFname           simple.paren.typedef.declarator member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration";)
		EXECUTE_DEBUG(cerr << " for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $2,
				    current_access_token);

		if ($3 != NULL)
		    $2->set_pure_virtual();
			
		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $2);
	    } 
        | declaration.specifier simple.paren.typedef.declarator member.pure.opt
            {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration qualifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $2);
		
		if ($3 != NULL)
		    $2->set_pure_virtual();

		/*
		 * Create a declaration object indicating the base type of this
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $2);
	    } 
        | member.conflict.paren.declaring.item
        ;

/* 
 * The following still conflicts with constructors-
 * member.conflict.paren.declaring.item:
 *      TYPEDEFname           paren.typedef.declarator     member.pure.opt
 *      | declaration.specifier paren.typedef.declarator     member.pure.opt
 *      ;
 * so paren.typedef.declarator is expanded inline to get...
 */

member.conflict.paren.declaring.item:
        TYPEDEFname   indirect.or.reference 
                '(' simple.paren.typedef.declarator ')' member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $4->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		$4->add_type(new type(new type_modifier($2)));

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $4);
			
		if ($6 != NULL)
		    $4->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $4);
	    } 
        | TYPEDEFname pointer.operator
                '(' simple.paren.typedef.declarator ')' member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $4->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		$4->add_type($2);

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $4);
			
		if ($6 != NULL)
		    $4->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $4);
	    } 
        | TYPEDEFname indirect.or.reference
                '(' TYPEDEFname ')'                     member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $4;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();

		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($4, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($4);
		
		declarator *typedef_declarator_ptr = 
		    new declarator(id_ptr, new type(new type_modifier($2)));

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(),
				    typedef_declarator_ptr);
			
		if ($6 != NULL)
		    typedef_declarator_ptr->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr,
					      typedef_declarator_ptr);
	    } 
        | TYPEDEFname pointer.operator
                '(' TYPEDEFname ')'                     member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $4;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();

		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($4, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($4);
		
		declarator *typedef_declarator_ptr = new declarator(id_ptr,$2);

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(),
				    typedef_declarator_ptr);
			
		if ($6 != NULL)
		    typedef_declarator_ptr->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr,
					      typedef_declarator_ptr);
	    } 
        | TYPEDEFname indirect.or.reference
                 paren.typedef.declarator               member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		$3->add_type(new type(new type_modifier($2)));

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $3);
			
		if ($4 != NULL)
		    $3->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $3);
	    } 
        | TYPEDEFname pointer.operator
                 paren.typedef.declarator               member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		$3->add_type($2);

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $3);
			
		if ($4 != NULL)
		    $3->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $3);
	    } 
        | declaration.specifier indirect.or.reference
                '(' simple.paren.typedef.declarator ')' member.pure.opt
	    {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $4->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		$4->add_type(new type(new type_modifier($2)));

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $4);
			
		if ($6 != NULL)
		    $4->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $4);
	    } 
        | declaration.specifier pointer.operator
                '(' simple.paren.typedef.declarator ')' member.pure.opt
	    {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $4->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		$4->add_type($2);

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $4);
			
		if ($6 != NULL)
		    $4->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $4);
	    } 
        | declaration.specifier indirect.or.reference
                '(' TYPEDEFname ')'                     member.pure.opt
	    {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $4;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($4, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($4);
		
		declarator *typedef_declarator_ptr = 
		    new declarator(id_ptr, new type(new type_modifier($2)));

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), typedef_declarator_ptr);
			
		if ($6 != NULL)
		    typedef_declarator_ptr->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, typedef_declarator_ptr);
	    } 
        | declaration.specifier pointer.operator
                '(' TYPEDEFname ')'                     member.pure.opt
	    {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $4;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($4, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($4);
		
		declarator *typedef_declarator_ptr = new declarator(id_ptr,$2);

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), typedef_declarator_ptr);
			
		if ($6 != NULL)
		    typedef_declarator_ptr->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, typedef_declarator_ptr);
	    } 
        | declaration.specifier indirect.or.reference
                paren.typedef.declarator                member.pure.opt
	    {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		$3->add_type(new type(new type_modifier($2)));

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $3);
			
		if ($4 != NULL)
		    $3->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $3);
	    } 
        | declaration.specifier pointer.operator
                paren.typedef.declarator                member.pure.opt
	    {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		$3->add_type($2);

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $3);
			
		if ($4 != NULL)
		    $3->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $3);
	    } 

        | member.conflict.paren.postfix.declaring.item
        ;

/* 
 * but we still have the following conflicts with constructors-
 * member.conflict.paren.postfix.declaring.item:
 *     TYPEDEFname           paren.postfix.typedef.declarator member.pure.opt
 *     | declaration.specifier paren.postfix.typedef.declarator member.pure.opt
 *     ;
 * so we expand paren.postfix.typedef inline and get...
 */

member.conflict.paren.postfix.declaring.item:
        TYPEDEFname     '(' paren.typedef.declarator ')'
                                                           member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $3);
			
		if ($5 != NULL)
		    $3->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $3);
	    } 
        | TYPEDEFname   '(' simple.paren.typedef.declarator
                        postfixing.abstract.declarator ')' member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();

		if ($3->declarator_type == NULL)
		    $3->declarator_type = $4;
		else
		{
		    $4->add_type($3->declarator_type);
		    $3->declarator_type = $4;
		}

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $3);
			
		if ($6 != NULL)
		    $3->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $3);
	    } 
        | TYPEDEFname   '(' TYPEDEFname
                        postfixing.abstract.declarator ')' member.pure.opt
	    {
		/*
		 * The first TYPEDEFname contains the name of a typedefed 
		 * identifier or of an aggregate(class, struct enum or union).
		 * This must be looked up in the symbol table and used to 
		 * build the type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $3 << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($3, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($3);
		
		declarator *declarator_ptr = new declarator(id_ptr, $4);

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(),
				    declarator_ptr);
			
		if ($4 != NULL)
		    declarator_ptr->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$=new variable_declaration(typedef_type_ptr, declarator_ptr);
	    } 
        | TYPEDEFname   '(' paren.typedef.declarator ')'
                        postfixing.abstract.declarator     member.pure.opt
	    {
		/*
		 * TYPEDEFname contains the name of a typedefed identifier or
		 * of an aggregate(class, struct enum or union).  This must
		 * be looked up in the symbol table and used to build the
		 * type for the identifier being declared.
		 */

		EXECUTE_DEBUG(cerr << "Typedefed type declaration for ";)
		EXECUTE_DEBUG(cerr << "identifier: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		if ($3->declarator_type == NULL)
		    $3->declarator_type = $5;
		else
		    $3->declarator_type->add_type($5);

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $3);
			
		if ($6 != NULL)
		    $3->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration(typedef_type_ptr, $3);
	    } 

        | declaration.specifier '(' paren.typedef.declarator ')' 
                                                           member.pure.opt
	    {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $3);
			
		if ($5 != NULL)
		    $3->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $3);
	    } 
        | declaration.specifier '(' simple.paren.typedef.declarator
                        postfixing.abstract.declarator ')' member.pure.opt
	    {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		if ($3->declarator_type == NULL)
		    $3->declarator_type = $4;
		else
		{
		    $4->add_type($3->declarator_type);
		    $3->declarator_type = $4;
		}

		general_declaration($1->duplicate(), $3);
			
		if ($6 != NULL)
		    $3->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $3);
	    } 
        | declaration.specifier '(' TYPEDEFname
                        postfixing.abstract.declarator ')' member.pure.opt
	    {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $3 << " in class declaration.\n";)

		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($3, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($3);
		
		declarator *declarator_ptr = new declarator(id_ptr, $4);

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), declarator_ptr);
			
		if ($4 != NULL)
		    declarator_ptr->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.  Since the
		 * base type is a typedef name create a new base type object
		 * for this name and store it in the declarator.
		 */

		$$ = new variable_declaration($1, declarator_ptr);
	    } 
        | declaration.specifier '(' paren.typedef.declarator ')' 
                        postfixing.abstract.declarator     member.pure.opt
	    {
		/*
		 * A declaration.specifier can contain storage classes, the 
		 * keywords typedef, volatile and const and a base type. 
		 */

		EXECUTE_DEBUG(cerr << "declaration specifier declaration for";)
		EXECUTE_DEBUG(cerr << " identifier: ";)
		EXECUTE_DEBUG(cerr << $3->declarator_id->id_string;)
		EXECUTE_DEBUG(cerr << " in class declaration.\n";)

		if ($3->declarator_type == NULL)
		    $3->declarator_type = $5;
		else
		{
		    $5->add_type($3->declarator_type);
		    $3->declarator_type = $5;
		}

		/*
		 * Call general_declaration on a duplicate of the type.  This 
		 * is done because the canonical type that is associated with 
		 * the identifier may have a  different form than the type to 
		 * be printed with the declarator record.
		 */

		general_declaration($1->duplicate(), $3);
			
		if ($6 != NULL)
		    $3->set_pure_virtual();

		/*
		 * Create a declaration object indicating the type of this 
		 * declaration and the identifier being declared.
		 */

		$$ = new variable_declaration($1, $3);
	    } 
        ;


member.pure.opt:
        /* nothing */
            {
		$$ = NULL;
	    }
        | '=' OCTALconstant /* C++, not ANSI C */ /* Pure function*/
            {
		// Should make sure the OCTALconstant is 0. - ecl.
		$$ = new constant($2, INT, OCTALconstant);
	    }
        ;

/******************************************************************************
 *
 *				 Bit Field Declarations
 *
 *****************************************************************************/

/*
 * Note that bit field names cannot be parenthesized in C++  (due 
 * to  ambiguities),  and  hence this part of the grammar is simpler 
 * than ANSI C. :-) 
 */

bit.field.declarator:
        bit.field.identifier.declarator
        | TYPEDEFname ':' constant.expression
            {
		$$ = new declarator(new identifier($1, IDENTIFIER));
		$$->bit_field_size_expression = $3;
	    }
        ;

/* 
 * The actions taken in the "{}" above and below are probably not 
 * really required (re: putting the symbol into the symbol table  as 
 * part  of  the  struct), as there is (I think) no way to reference 
 * them until the struct is complete.  Hence there is no  real  risk 
 * that  the constant.expression would make reference to them (e.g., 
 * "name : sizeof  name").   For  cleanliness,  we  will  leave  the 
 * actions in this form. 
 *
 * They cause trouble for me so I am removing them. - ecl.
 */

bit.field.identifier.declarator:
        ':' constant.expression
            {
		$$ = new declarator((identifier *)NULL);
		$$->bit_field_size_expression = $2;
	    }
        | IDENTIFIER ':' constant.expression
            {
		$$ = new declarator(new identifier($1, IDENTIFIER));
		$$->bit_field_size_expression = $3;
	    }
        ;

/******************************************************************************
 *
 *			  Enum Declarations and Definitions
 *
 *****************************************************************************/

enum.name:
        ENUM '{' enumerator.list '}'
            {
		/*
		 * An unnamed enum definition is being parsed.  
		 *
		 * enumerator.list - A declarator list containing the enum
		 * 		     definitions.
		 */
		
		/*
		 * Create an unnamed identifier to be used in the enum
		 * definition.
		 */

		identifier *id_ptr = new identifier(NULL, ENUM);

		// Create an enum definition object for this enum.
		enum_definition *enum_definition_ptr = 
		    new enum_definition(id_ptr, $3);

		// Set the type of the identifier object.
		base_type *base_type_ptr = new base_type(TYPEDEFname, id_ptr);
		base_type_ptr->typedef_type_ptr = 
		    new type(new base_type(ENUM, id_ptr));
		id_ptr->type_ptr = new type(base_type_ptr);

		// Set the enum definition field in the identifier object.
		id_ptr->enum_definition_ptr = enum_definition_ptr;

		// Return an enum definition object.
		$$ = enum_definition_ptr;
	    }
        | ENUM identifier.or.typedef.name '{' enumerator.list '}'
            {
		/*
		 * An enum definition is being parsed.  
		 *
		 * identifier.or.typedef.name - string representation of the
		 *                              enum name.
		 * 
		 * enumerator.list - A declarator list containing the enum
		 * 		     definitions.
		 */

		/*
		 * Enter the enum name in the symbol table. If the name is 
		 * already in there log an error.
		 */

		identifier *id_ptr = scope_stack.insert($2, ENUM);
		if (id_ptr == NULL)
		{
		    /*
		     * The enum identifier is being redefined,log the error
		     * and ignore the previous definition.
		     */

		    cerr << program_name << ": line " << linenum;
		    cerr << " Identifier: " << $2;
		    cerr << " being redefined! Ignoring previous ";
		    cerr << "definition or declaration!\n";
		    id_ptr = scope_stack.remove(id_ptr);
		    id_ptr = scope_stack.insert($2, ENUM);
		}

		// Create an enum definition object for this enum.
		enum_definition *enum_definition_ptr = 
		    new enum_definition(id_ptr, $4);
		
		// Set the type of the enum identifier if neccesary.
		if (id_ptr->type_ptr == NULL)
		{
		    // Create the new base_type object.
		    base_type *base_type_ptr = 
			new base_type(TYPEDEFname, id_ptr);
		    base_type_ptr->typedef_type_ptr = 
			new type(new base_type(ENUM, id_ptr));
		    id_ptr->type_ptr = new type(base_type_ptr);
		}

		// Set the enum definition field in the identifier object.
		id_ptr->enum_definition_ptr = enum_definition_ptr;

		// Return the enum defintion.
		$$ = enum_definition_ptr;
	    }
        | ENUM identifier.or.typedef.name
            {
		/*
		 * An enum definition is being parsed.  
		 *
		 * identifier.or.typedef.name - string representation of the
		 *                              enum name.
		 */

		// An enum is being referenced.
		identifier *id_ptr = scope_stack.lookup($2, ENUM);
		if (id_ptr == NULL)
		{
		    /*
		     * An enum cannot be forward defined, thus this is an
		     * error. Log the error.
		     */

		    cerr << program_name << ": line " << linenum;
		    cerr << " Forward declaration of enum : " << $2;
		    cerr << "!\n";
		    id_ptr = scope_stack.insert($2, ENUM);
		}

		//Create an enum definition object for this enum.
		enum_definition *enum_definition_ptr = 
		    new enum_definition(id_ptr);

		// Set the type of the enum identifier if neccesary.
		if (id_ptr->type_ptr == NULL)
		{
		    // Create the new base_type object.
		    base_type *base_type_ptr = 
			new base_type(TYPEDEFname, id_ptr);
		    base_type_ptr->typedef_type_ptr = 
			new type(new base_type(ENUM, id_ptr));
		    id_ptr->type_ptr = new type(base_type_ptr);
		}

		// There is no body with this enum.
		enum_definition_ptr->print_body = FALSE;

		// Return the enum definition.
		$$ = enum_definition_ptr;
	    }
        ;

enumerator.list:
        enumerator.list.no.trailing.comma
        | enumerator.list.no.trailing.comma ',' /* C++, not ANSI C */
        ;

/*
 * Note  that we do not need to rush to add an enumerator to the 
 * symbol table until *AFTER* the  enumerator.value.opt  is  parsed.  
 * The  enumerated  value  is  only in scope AFTER its definition is 
 * complete.  Hence the following is legal: "enum {a, b=a+10};"  but 
 * the  following is (assuming no external matching of names) is not 
 * legal: "enum {c, d=sizeof(d)};" ("d" not defined when sizeof  was 
 * applied.)  This  is  notably  contrasted  with declarators, which 
 * enter scope as soon as the declarator is complete. 
 */

enumerator.list.no.trailing.comma:
        identifier.or.typedef.name enumerator.value.opt
            {
		/*
		 * Enter each name in the symbol table as an identifier of
		 * type int.  Return the declarator for the identifier.
		 */

		identifier *id_ptr = scope_stack.insert($1, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_ptr = id_redeclared_error($1);

		// Set the type to be type int.
		id_ptr->type_ptr = new type(new base_type(INT));
		
		// Create a new declarator object for this enummerator.
		declarator *new_declarator = new declarator(id_ptr);

		// If there is an initializer set the declarator objects field.
		if ($2 != NULL)
		    new_declarator->initializers = new initializer($2);

		// Return the declarator object for this enummerator.
		$$ = new declarator_list(new_declarator);
	    }
        | enumerator.list.no.trailing.comma ',' 
               identifier.or.typedef.name enumerator.value.opt
            {
		/*
		 * Enter each name in the symbol table as an identifier of
		 * type int.  Return the declarator for the identifier.
		 */

		identifier *id_ptr = scope_stack.insert($3, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_ptr = id_redeclared_error($3);
		
		// Set the type to be type int.
		id_ptr->type_ptr = new type(new base_type(INT));
		
		// Create a new declarator object for this enummerator.
		declarator *new_declarator = new declarator(id_ptr);

		// If there is an initializer set the declarator objects field.
		if ($4 != NULL)
		    new_declarator->initializers = new initializer($4);

		// Add the declarator for this enummerator to the list.
		$1->add_declarator(new_declarator);

		// Return the list of enummerators declarators.
		$$ = $1;
	    }
        ;

enumerator.value.opt:
        /* Nothing */
            {
		$$ = NULL;
	    }
        | '=' constant.expression
            {
		$$ = $2;
	    }
        ;

/******************************************************************************
 *
 *				Parameters
 *
 *****************************************************************************/

/*
 * We special case the lone type.name which has no storage  class 
 * (even  though  it should be an example of a parameter.type.list). 
 * This helped to disambiguate type-names in parenthetical casts.
 */

parameter.type.list: 
        '(' ')'                             type.qualifier.list.opt
            {
		if ($3 != NULL)
		    $$ = new parameter_list(new parameter(), $3);
		else
		    $$ = new parameter_list(new parameter());		    
	    }
        | '(' type.name ')'                 type.qualifier.list.opt
            {
		if ($4 != NULL)
		    $$ = new parameter_list(new parameter($2), $4);
		else
		    $$ = new parameter_list(new parameter($2));
	    }
        | '(' type.name initializer ')'     type.qualifier.list.opt
            {
		if ($5 != NULL)
		    $$ = new parameter_list(new parameter($2, $3), $5);
		else
		    $$ = new parameter_list(new parameter($2, $3));
	    }
        | '(' named.parameter.type.list ')' type.qualifier.list.opt
            {
		$$ = $2;
		if ($4 != NULL)
		    $$->type_qualifier_list_ptr = $4;
	    }
        ;
    
named.parameter.type.list:  /* WARNING: excludes lone type.name*/
        parameter.list
        | parameter.list comma.opt.ellipsis
            {
		$1->add_parameter($2);
		$$ = $1;
	    }
        | type.name comma.opt.ellipsis
            {
		$$ = new parameter_list(new parameter($1));
		$$->add_parameter($2);
	    }
        | type.name initializer comma.opt.ellipsis
            {
		$$ = new parameter_list(new parameter($1, $2));
		$$->add_parameter($3);
	    }
        | ELLIPSIS
            {
		$$ = new parameter_list(new parameter($1));
	    }
        ;


comma.opt.ellipsis:
        ELLIPSIS    
            {
		$$ = new parameter($1);
	    }
        | ',' ELLIPSIS
            {
		$$ = new parameter($2);
	    }
        ;

parameter.list: 
        named.parameter.declaration
            {
		$$ = new parameter_list($1);
	    }
        | named.parameter.declaration initializer
            {
		$1->parameter_initializer = $2;
		$$ = new parameter_list($1);
	    }
        | type.name             ',' parameter.declaration
            {
		$$ = new parameter_list(new parameter($1));
		$$->add_parameter($3);
	    }
        | type.name initializer ',' parameter.declaration
            {
		$$ = new parameter_list(new parameter($1, $2));
		$$->add_parameter($4);
	    }
        | parameter.list ',' parameter.declaration
            {
		$1->add_parameter($3);
		$$ = $1;
	    }
        ;

/*
 * There  is  some very subtle disambiguation going on here.  Do 
 * not be tempted to make further use of the following production in 
 * parameter.list, or else the conflict count will grow  noticeably.  
 * Specifically,  the  next  set  of  rules  has already been inline 
 * expanded for the first parameter in a parameter.list to support a 
 * deferred disambiguation. 
 */
     
parameter.declaration:
        type.name
            {
		$$ = new parameter($1);
	    }
        | type.name initializer  
            {
		$$ = new parameter($1, $2);
	    }
        | named.parameter.declaration
        | named.parameter.declaration initializer 
            {
		$1->parameter_initializer = $2;
		$$ = $1;
	    }
        ;

/*
 * One big thing implemented here is that a TYPEDEFname CANNOT be 
 * redeclared when we don't have declaration.specifiers! Notice that 
 * when we do use a TYPEDEFname based declarator, only the "special" 
 * (non-ambiguous  in  this  context)  typedef.declarator  is  used. 
 * Everything else that is "missing" shows up as a type.name. 
 */

named.parameter.declaration: /*have names or storage classes */
        declaration.specifier
            {
		$$ = new parameter(new variable_declaration($1));
	    }
        | declaration.specifier abstract.declarator
            {
	     $$=new parameter(new variable_declaration($1,new declarator($2)));
	    }
        | declaration.specifier identifier.declarator
            {
		$$ = new parameter(new variable_declaration($1, $2));
	    }
        | declaration.specifier parameter.typedef.declarator
            {
		$$ = new parameter(new variable_declaration($1, $2));
	    }

        | declaration.qualifier.list 
            {
		$$ = new parameter(new variable_declaration($1));
	    }
        | declaration.qualifier.list abstract.declarator
            {
	     $$=new parameter(new variable_declaration($1,new declarator($2)));
	    }
        | declaration.qualifier.list identifier.declarator
            {
		$$ = new parameter(new variable_declaration($1, $2));
	    }
        | type.specifier identifier.declarator
            {
		$$ = new parameter(new variable_declaration($1, $2));
	    }
        | type.specifier parameter.typedef.declarator
            {
		$$ = new parameter(new variable_declaration($1, $2));
	    }

        | basic.type.name identifier.declarator
            {
		$$ = new parameter(new variable_declaration(new type($1), $2));
	    }
        | basic.type.name parameter.typedef.declarator
            {
		$$ = new parameter(new variable_declaration(new type($1), $2));
	    }

        | TYPEDEFname identifier.declarator
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($1);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();

		$$ = new parameter(new variable_declaration(typedef_ptr, $2));
	    }
        | TYPEDEFname parameter.typedef.declarator
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($1);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();

		$$ = new parameter(new variable_declaration(typedef_ptr, $2));
	    }

        | type.qualifier.list identifier.declarator
            {
		$$ = new parameter(new variable_declaration($1, $2));
	    }
        ;
    
identifier.or.typedef.name:
        IDENTIFIER
        | TYPEDEFname 
        ;

type.name:
        type.specifier
            {
		$$ = new variable_declaration($1);
	    }
        | basic.type.name 
            {
		$$ = new variable_declaration(new type($1));
	    }
        | TYPEDEFname 
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($1);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();

		$$ = new variable_declaration(typedef_ptr);
	    }
        | type.qualifier.list 
            {
		$$ = new variable_declaration($1);
	    }

        | type.specifier abstract.declarator
            {
		$$ = new variable_declaration($1, new declarator($2));
	    }
        | basic.type.name abstract.declarator
            {
		$$ = new variable_declaration(new type($1),new declarator($2));
	    }
        | TYPEDEFname abstract.declarator
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($1);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();

		$$ = new variable_declaration(typedef_ptr, new declarator($2));
	    }
        | type.qualifier.list abstract.declarator
            {
		$$ = new variable_declaration($1, new declarator($2));
	    }
        ;

/******************************************************************************
 *
 *				 Initializers
 *
 *****************************************************************************/

initializer.opt:
        /* nothing */
        {    
	    $$ = NULL;
	}
        | initializer
        ;

initializer:
        '=' initializer.group
        {    
	    $$ = $2;
	}
        ;

initializer.group:        
        '{' initializer.list '}'
            {
		$$ = new initializer($2);
	    }
        | '{' initializer.list ',' '}'
            {
		$$ = new initializer($2);
	    }
        | assignment.expression
            {    
		$$ = new initializer($1);
	    }
        ;

initializer.list:
        initializer.group
            {
		$$ = new initializer_list($1);
	    }
        | initializer.list ',' initializer.group
            {
		$1->add_initializer($3);
		$$ = $1;
	    }
        ;


/******************************************************************************
 *
 *				 Statements
 *
 *****************************************************************************/

statement:
        labeled.statement
        | compound.statement
        | expression.statement
        | selection.statement
        | iteration.statement
        | jump.statement
        ;

labeled.statement:
        identifier.or.typedef.name ':' statement
            {    
		// A label is being defined.  Insert it in the current scope.
		identifier *temp_id_ptr = scope_stack.insert($1, LABEL);
		if (temp_id_ptr == NULL)
		{
		    cerr << "Error : Label string : " <<temp_id_ptr->id_string;
		    cerr << " is already defined!\n";
		    temp_id_ptr = scope_stack.lookup($1);
		}

		$$ = new labeled_statement(LABEL, temp_id_ptr, $3);
	    }
        | CASE constant.expression ':' statement
            {    
		$$ = new labeled_statement(CASE, $2, $4);
	    }
        | DEFAULT ':' statement
            {    
		$$ = new labeled_statement(DEFAULT, NULL, $3);
	    }
        ;

/*
 * I sneak declarations into statement.list to support C++.  The 
 * grammar is a little clumsy this  way,  but  the  violation  of  C 
 * syntax is heavily localized.
 */

start.compound.statement:
	'{' 
            {
		EXECUTE_DEBUG(cerr << "New scope for a compound statement!\n";)

		/*
		 * Each compound statement introduces a new scope, unless it
		 * is for the body of a function, then it must use the scope
		 * with the parameters in it.  Allocate a new symbol table and 
		 * make this new symbol table be the current scope.  If the 
		 * global parameter_table is set, use this as the new symbol 
		 * table.
		 */

		symbol_table *temp_table;

		if (parameter_table != NULL)
		{
		    EXECUTE_DEBUG(cerr << "Using the parameter table!\n";)
		    temp_table = parameter_table;
		    parameter_table = NULL;
		}
		else
		    temp_table = new symbol_table;

		scope_stack.push(temp_table);

		$$ = temp_table;
	    }
	;

compound.statement:
        start.compound.statement '}' 
            {
		EXECUTE_DEBUG(cerr << "Empty compound statement!\n";)
		EXECUTE_DEBUG(cerr << "Close the scope!\n";)
		/*
		 * An empty compound statement has been parsed.  Build the
		 * compound statement object and reset the scope to the
		 * surrounding scope.
		 */

		compound_statement *temp_stmt = new compound_statement($1);

		TABLE_DEBUG(scope_stack.dump_top();)
		scope_stack.pop();

		$$ = temp_stmt;
	    }
        | start.compound.statement declaration.list '}'
            {
	        EXECUTE_DEBUG(cerr << "Compound statement with initial ";)
		EXECUTE_DEBUG(cerr << "declarations!\nClose the scope!\n";)
		EXECUTE_DEBUG($1->print();)
		//Build a compound statement object using the new symbol table.
		compound_statement *temp_stmt = new compound_statement($1);

		// Add the declarations to the compound statement object.
		temp_stmt->declarations = $2;

		/*
		 * A compound statement has its own scope.  Close this scope
		 * and reset the scope to be the scope surrounding this 
		 * compound statement.
		 */

		TABLE_DEBUG(scope_stack.dump_top();)
		scope_stack.pop();

		$$ = temp_stmt;
	    }
        | start.compound.statement statement.list '}'
            {
	        EXECUTE_DEBUG(cerr << "Compound statement with statements!\n";)
		EXECUTE_DEBUG(cerr << "Close the scope!\n";)
		EXECUTE_DEBUG($1->print();)
		//Build a compound statement object using the new symbol table.
		compound_statement *temp_stmt = new compound_statement($1);

		// Add the statements to the compound statement object.
		temp_stmt->statements = $2;

		/*
		 * A compound statement has its own scope.  Close this scope
		 * and reset the scope to be the scope surrounding this 
		 * compound statement.
		 */

		TABLE_DEBUG(scope_stack.dump_top();)
		scope_stack.pop();

		$$ = temp_stmt;
	    }
        | start.compound.statement declaration.list statement.list '}'
            {
	        EXECUTE_DEBUG(cerr << "Compound statement with initial ";)
		EXECUTE_DEBUG(cerr << "declarations and statements!\n";)
		EXECUTE_DEBUG(cerr << "Close the scope!\n";)
		EXECUTE_DEBUG($1->print();)
		//Build a compound statement object using the new symbol table.
		compound_statement *temp_stmt = new compound_statement($1);

		// Add the declarations to the compound statement object.
		temp_stmt->declarations = $2;

		// Add the statements to the compound statement object.
		temp_stmt->statements   = $3;

		/*
		 * A compound statement has its own scope.  Close this scope
		 * and reset the scope to be the scope surrounding this 
		 * compound statement.
		 */

		TABLE_DEBUG(scope_stack.dump_top();)
		scope_stack.pop();

		$$ = temp_stmt;
	    }
        ;

declaration.list:
        declaration
            {
		EXECUTE_DEBUG(cerr << "Creating a declaration list!\n";)
		/*
		 * Declarations are treated as statements in c++, therefore
		 * build a statement list out of this declaration.
		 */

		$$ = new statement_list($1);
	    }
        | declaration.list declaration
            {
	      EXECUTE_DEBUG(cerr<<"Adding declaration to declaration list!\n";)
		// Add the declaration to the declaration (statement) list.
		$1->add_statement($2);
		$$ = $1;
	    }
        ;

statement.list:
        statement
            {
		EXECUTE_DEBUG(cerr << "Creating a statement list!\n";)
		// Build a statement list.
		$$ = new statement_list($1);
	    }
        | statement.list statement
            {
		EXECUTE_DEBUG(cerr << "Adding statement to statement list!\n";)
		// Add the statement to the statement list.
		$1->add_statement($2);
		$$ = $1;
	    }
        | statement.list declaration /* C++, not ANSI C */
            {
		EXECUTE_DEBUG(cerr<<"Adding declaration to statement list!\n";)
		/*
		 * Declarations are treated as statements in c++, therefore
		 * add this declaration to the statement list.
		 */
		$1->add_statement($2);
		$$ = $1;
	    }
        ;

expression.statement:
        expression.opt ';'
            {
		$$ = new expression_statement($1);
	    }
        ;

selection.statement:
        IF '(' expression ')' statement
            {    
		$$ = new selection_statement(IF, $3, $5, NULL);
	    }
        | IF '(' expression ')' statement ELSE statement
            {    
		$$ = new selection_statement(IF, $3, $5, $7);
	    }
        | SWITCH '(' expression ')' statement
            {    
		$$ = new selection_statement(SWITCH, $3, $5, NULL);
	    }
        ;

iteration.statement:
        WHILE '(' expression ')' statement
            {
		$$ = new while_statement(WHILE, $3, $5);
	    }
        | DO statement WHILE '(' expression ')' ';'
            {
		$$ = new while_statement(DO, $5, $2);
	    }
        | FOR '(' expression.opt ';' expression.opt ';'
                expression.opt ')' statement
            {
		/*
		 * To be consistant, make a statement out of the first
		 * expression and store it in the for statement object as 
		 * a statement.
		 */

		statement *temp_stmt_ptr = new expression_statement($3);

		$$ = new for_statement(temp_stmt_ptr, $5, $7, $9);
	    }
        | FOR '(' declaration        expression.opt ';'
                expression.opt ')' statement  /* C++, not ANSI C */
            {
		$$ = new for_statement($3, $4, $6, $8);
	    }
        ;

jump.statement:
        GOTO identifier.or.typedef.name ';'
            {
		$$ = new jump_statement(GOTO, new identifier($2, IDENTIFIER));
	    }
        | CONTINUE ';'
            {    
		$$ = new jump_statement(CONTINUE, NULL);
	    }
        | BREAK ';'
            {    
		$$ = new jump_statement(BREAK, NULL);
	    }
        | RETURN expression.opt ';'
            {    
		$$ = new jump_statement(RETURN, $2);
	    }
        ;


/******************************************************************************
 *
 *			 External Definitions
 *
 *****************************************************************************/

translation.unit:
        external.definition
            {
		$$ = new definition_list($1);
	    }
        | translation.unit external.definition
            {
		$1->add_definition($2);
		$$ = $1;
	    }
        ;

external.definition:
        function.declaration
	    {
		$$ = $<definition_ptr>1;
	    }
        | function.definition
	    {
		$$ = $<definition_ptr>1;
	    }
        | declaration
            {    
		$$ = new external_definition($1);
	    }
        | linkage.specifier function.declaration     /* C++, not ANSI C*/
            {    
		$$ = new external_definition($2, $1);
	    }
        | linkage.specifier function.definition      /* C++, not ANSI C*/
            {    
		$$ = new external_definition($2, $1);
	    }
        | linkage.specifier declaration              /* C++, not ANSI C*/
            {    
		$$ = new external_definition($2);
		((external_definition *)$$)->linkage_specifier = $1
	    }
        | linkage.specifier '{' translation.unit '}' /* C++, not ANSI C*/
            {    
		$$ = new external_definition($3, $1);
	    }
        | linkage.specifier '{' '}' /* C++, not ANSI C*/
            {    
		$$ = new external_definition(new definition_list(NULL), $1);
	    }
        ;

linkage.specifier:
        EXTERN STRINGliteral
	    {
		$$ = new constant($2, STRINGliteral, STRINGliteral);
	    }
        ;

/******************************************************************************
 *
 *     		  	Function Definitions and Declarations
 *
 *****************************************************************************/

/*
 * Note that declaration.specifiers are left out of the following 
 * function declarations.  This is something of an anomaly in C, but 
 * an  unfortunately  common  coding practice.  It is also sometimes 
 * necessary in C++, in instances where no  return  type  should  be 
 * specified (e.g., a conversion operator).
 */

function.declaration:
        identifier.declarator ';' /* semantically verify it is a function */
            {
		EXECUTE_DEBUG(cerr << "Function declaration for function ";)
		EXECUTE_DEBUG(cerr << " without return type named : ";)
		EXECUTE_DEBUG(cerr << $1->declarator_id->id_string << "!\n";)

		// By default this function returns type int.
		type *base_type_ptr = new type(new base_type(INT));
		
		general_declaration(base_type_ptr, $1);

		function_definition *temp_func_def = 
		    new function_definition(NULL, $1, NULL);

		$$ = temp_func_def;
	    }
        ;
        
function.definition:
        new.function.definition
        | constructor.function.definition
        ;

new.function.definition:
                                     identifier.declarator 
            {
		/*
		 * A function definition without a return type is being parsed.
		 *
		 * identifier.declarator - the function name including the
		 * 			   parameters.
		 *
		 * compound.statement - a statement list containing the
		 *			statements in the function body.
		 */

		/*
		 * For member function definitions the corresponding class
		 * members must be in scope.  If this is a member function 
		 * definition the  identifier.declarator will have a scope_list
		 * entry. Call the member function set_scope for the scope_list
		 * entry to setup the correct scope.
		 */

		if ($1->declarator_id->scope_list != NULL)
		{
		    class_definition *class_def_ptr = 
			$1->declarator_id->scope_list->get_last();
		    if (class_def_ptr->mentat_class_flag)
		    {
			if (mentat_class == NULL)
			{
			    mentat_class = class_def_ptr;

			    // Enter the SELF record in the symbol table.
			    type *self_type_ptr = class_def_ptr->
				identifier_ptr->type_ptr->duplicate();
			    identifier *self_id_ptr = 
				new identifier("SELF",IDENTIFIER);
			    self_type_ptr->expand_typedefs();
			    self_id_ptr->type_ptr = self_type_ptr;
			    self_id_ptr = global_table.insert(self_id_ptr);
			}
			else
			    if (mentat_class != class_def_ptr)
			    {
				// Log an error,more than one mentat class with
				// member function definitions.
				cerr << "Multiple Mentat classes with member";
				cerr << " function definitons seen!\n";
				exit(-1);
			    }			    
		    }
		    
		    current_class.push(class_def_ptr);

		    $1->declarator_id->scope_list->set_scope();
		}
		
		/*
		 * The identifier declarator is a function name with
		 * parameter declarations.  Set up a symbol table containing
		 * the parameters and use this symbol table when parsing the
		 * function body.  
		 */

		$1->set_parameter_symbol_table();

		if ($1->declarator_id->id_string != NULL && 
		    strcmp($1->declarator_id->id_string, "main") == 0)
		    main_flag = TRUE;
	    }
                                                            compound.statement
            {
		EXECUTE_DEBUG(cerr << "Function definition for function ";)
		EXECUTE_DEBUG(cerr << " without return type named : ";)
		EXECUTE_DEBUG(cerr << $1->declarator_id->id_string << "!\n";)

		// By default this function returns type int.
		type *base_type_ptr = new type(new base_type(INT));
		
		/*
		 * Call general_declaration to enter this function definition
		 * in the symbol table.  Pass a duplicate of the type because 
		 * the canonical type entered with the identifier in the symbol
		 * table may have a different form than the type stored with 
		 * the declarator object.  Do not enter overloaded operators or
		 * conversion functions in the symbol table.
		 */

		if ($1->declarator_id->operator_flag != TRUE)
		  general_declaration(base_type_ptr->duplicate(),$1,NONE,TRUE);
		
		// Create a function definition object for this function.
		function_definition *function_definition_ptr = 
		    new function_definition(NULL, $1,(compound_statement *)$3);

		/*
		 * Set the definition field in the identifier being declared to
		 * point to the function definition object.
		 */

		$1->declarator_id->function_definition_ptr = 
		    function_definition_ptr;

		// If this is a member function reset the scope.
		if ($1->declarator_id->scope_list != NULL)
		{
		    current_class.pop();
		    $1->declarator_id->scope_list->reset_scope();
		}
		
		if (main_flag)
		{
		    if (main_function != NULL)
			// Log an error. Two declarations for main!!. - ecl
			;
		    main_function = function_definition_ptr;
		    main_flag = FALSE;
		}

		$$ = function_definition_ptr;
	    }
        | declaration.specifier      identifier.declarator 
            {
		/*
		 * A function definition is being parsed.
		 *
		 * declaration.specifier - can contain storage classes, the 
		 *			   keywords typedef, volatile and const
		 *			   and a base type.
		 *
		 * identifier.declarator - the function name including the
		 * 			   parameters.
		 *
		 * compound.statement - a statement list containing the
		 *			statements in the function body.
		 */

		/*
		 * For member function definitions the corresponding class
		 * members must be in scope.  If this is a member function 
		 * definition the  identifier.declarator will have a scope_list
		 * entry. Call the member function set_scope for the scope_list
		 * entry to setup the correct scope.
		 */

		if ($2->declarator_id->scope_list != NULL)
		{
		    class_definition *class_def_ptr = 
			$2->declarator_id->scope_list->get_last();
		    if (class_def_ptr->mentat_class_flag)
		    {
			if (mentat_class == NULL)
			{
			    mentat_class = class_def_ptr;

			    // Enter the SELF record in the symbol table.
			    type *self_type_ptr = class_def_ptr->
				identifier_ptr->type_ptr->duplicate();
			    identifier *self_id_ptr = 
				new identifier("SELF",IDENTIFIER);
			    self_type_ptr->expand_typedefs();
			    self_id_ptr->type_ptr = self_type_ptr;
			    self_id_ptr = global_table.insert(self_id_ptr);
			}
			else
			    if (mentat_class != class_def_ptr)
			    {
				// Log an error,more than one mentat class with
				// member function definitions.
				cerr << "Multiple Mentat classes with member";
				cerr << " function definitons seen!\n";
			    }			    
		    }
		    
		    current_class.push(class_def_ptr);

		    $2->declarator_id->scope_list->set_scope();
		}
		
		/*
		 * The identifier declarator is a function name with
		 * parameter declarations.  Set up a symbol table containing
		 * the parameters and use this symbol table when parsing the
		 * function body.  
		 */

		$2->set_parameter_symbol_table();

		if (strcmp($2->declarator_id->id_string, "main") == 0)
		    main_flag = TRUE;
	    } 
                                                           compound.statement
            {
		EXECUTE_DEBUG(cerr << "declaration specifier function";)
		EXECUTE_DEBUG(cerr << " definition for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Call general_declaration to enter this function definition
		 * in the symbol table.  Pass a duplicate of the type because 
		 * the canonical type entered with the identifier in the symbol
		 * table may have a different form than the type stored with
		 * the declarator object.
		 */

		general_declaration($1->duplicate(), $2, NONE, TRUE);
		
		// Create a function definition object for this function.
		function_definition *function_definition_ptr = 
		    new function_definition($1, $2, (compound_statement *)$4);

		/*
		 * Set the definition field in the identifier being declared to
		 * point to the function definition object.
		 */

		$2->declarator_id->function_definition_ptr = 
		    function_definition_ptr;

		// If this is a member function reset the scope.
		if ($2->declarator_id->scope_list != NULL)
		{
		    current_class.pop();
		    $2->declarator_id->scope_list->reset_scope();
		}

		if (main_flag)
		{
		    if (main_function != NULL)
			// Log an error. Two declarations for main!!. - ecl
			;
		    main_function = function_definition_ptr;
		    main_flag = FALSE;
		}

		$$ = function_definition_ptr;
	    } 
        | type.specifier             identifier.declarator 
            {
		/*
		 * A function definition is being parsed.
		 *
		 * type.specifier - contains at least one of the keywords const
		 *		    and volatile and a base type.
		 *
		 * identifier.declarator - the function name including the
		 * 			   parameters.
		 *
		 * compound.statement - a statement list containing the
		 *			statements in the function body.
		 */

		/*
		 * For member function definitions the corresponding class
		 * members must be in scope.  If this is a member function 
		 * definition the  identifier.declarator will have a scope_list
		 * entry. Call the member function set_scope for the scope_list
		 * entry to setup the correct scope.
		 */

		if ($2->declarator_id->scope_list != NULL)
		{
		    class_definition *class_def_ptr = 
			$2->declarator_id->scope_list->get_last();
		    if (class_def_ptr->mentat_class_flag)
		    {
			if (mentat_class == NULL)
			{
			    mentat_class = class_def_ptr;

			    // Enter the SELF record in the symbol table.
			    type *self_type_ptr = class_def_ptr->
				identifier_ptr->type_ptr->duplicate();
			    identifier *self_id_ptr = 
				new identifier("SELF",IDENTIFIER);
			    self_type_ptr->expand_typedefs();
			    self_id_ptr->type_ptr = self_type_ptr;
			    self_id_ptr = global_table.insert(self_id_ptr);
			}
			else
			    if (mentat_class != class_def_ptr)
			    {
				// Log an error,more than one mentat class with
				// member function definitions.
				cerr << "Multiple Mentat classes with member";
				cerr << " function definitons seen!\n";
			    }			    
		    }

		    current_class.push(class_def_ptr);
		    
		    $2->declarator_id->scope_list->set_scope();
		}
		
		/*
		 * The identifier declarator is a function name with
		 * parameter declarations.  Set up a symbol table containing
		 * the parameters and use this symbol table when parsing the
		 * function body.  
		 */

		$2->set_parameter_symbol_table();

		if (strcmp($2->declarator_id->id_string, "main") == 0)
		    main_flag = TRUE;
	    } 
                                                            compound.statement
            {
		EXECUTE_DEBUG(cerr << "type specifier function";)
		EXECUTE_DEBUG(cerr << " definition for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Call general_declaration to enter this function definition 
		 * in the symbol table.  Pass a duplicate of the type because 
		 * the canonical type entered with the identifier in the symbol
		 * table may have a different form than the type stored with 
		 * the declarator object.
		 */

		general_declaration($1->duplicate(), $2, NONE, TRUE);
		
		// Create a function definition object for this function.
		function_definition *function_definition_ptr = 
		    new function_definition($1, $2, (compound_statement *)$4);

		/*
		 * Set the definition field in the identifier being declared to
		 * point to the function definition object.
		 */

		$2->declarator_id->function_definition_ptr = 
		    function_definition_ptr;

		// If this is a member function reset the scope.
		if ($2->declarator_id->scope_list != NULL)
		{
		    current_class.pop();
		    $2->declarator_id->scope_list->reset_scope();
		}

		if (main_flag)
		{
		    if (main_function != NULL)
			// Log an error. Two declarations for main!!. - ecl
			;
		    main_function = function_definition_ptr;
		    main_flag = FALSE;
		}

		$$ = function_definition_ptr;
	    } 
        | basic.type.name            identifier.declarator 
            {
		/*
		 * A function definition is being parsed.
		 *
		 * basic.type.name - contains a base type.
		 *
		 * identifier.declarator - the function name including the
		 * 			   parameters.
		 *
		 * compound.statement - a statement list containing the
		 *			statements in the function body.
		 */

		/*
		 * For member function definitions the corresponding class
		 * members must be in scope.  If this is a member function 
		 * definition the  identifier.declarator will have a scope_list
		 * entry. Call the member function set_scope for the scope_list
		 * entry to setup the correct scope.
		 */

		if ($2->declarator_id->scope_list != NULL)
		{
		    class_definition *class_def_ptr = 
			$2->declarator_id->scope_list->get_last();
		    if (class_def_ptr->mentat_class_flag)
		    {
			if (mentat_class == NULL)
			{
			    mentat_class = class_def_ptr;

			    // Enter the SELF record in the symbol table.
			    type *self_type_ptr = class_def_ptr->
				identifier_ptr->type_ptr->duplicate();
			    identifier *self_id_ptr = 
				new identifier("SELF",IDENTIFIER);
			    self_type_ptr->expand_typedefs();
			    self_id_ptr->type_ptr = self_type_ptr;
			    self_id_ptr = global_table.insert(self_id_ptr);
			}
			else
			    if (mentat_class != class_def_ptr)
			    {
				// Log an error,more than one mentat class with
				// member function definitions.
				cerr << "Multiple Mentat classes with member";
				cerr << " function definitons seen!\n";
			    }			    
		    }
		    
		    current_class.push(class_def_ptr);

		    $2->declarator_id->scope_list->set_scope();
		}
		
		/*
		 * The identifier declarator is a function name with
		 * parameter declarations.  Set up a symbol table containing
		 * the parameters and use this symbol table when parsing the
		 * function body.  
		 */

		$2->set_parameter_symbol_table();

		if (strcmp($2->declarator_id->id_string, "main") == 0)
		    main_flag = TRUE;
	    } 
                                                          compound.statement
            {
		EXECUTE_DEBUG(cerr << "basic type name function";)
		EXECUTE_DEBUG(cerr << " definition for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Call general_declaration to enter this function definition 
		 * in the symbol table.Pass a duplicate of the type because the
		 * canonical type entered with the identifier in the symbol 
		 * table may have a different form than the type stored with 
		 * the declarator object.
		 */

		general_declaration(new type($1->duplicate()), $2, NONE, TRUE);
		
		// Create a function definition object for this function.
		function_definition *function_definition_ptr = 
		    new function_definition(new type($1), $2, 
					    (compound_statement *)$4);

		/*
		 * Set the definition field in the identifier being declared to
		 * point to the function definition object.
		 */

		$2->declarator_id->function_definition_ptr = 
		    function_definition_ptr;

		// If this is a member function reset the scope.
		if ($2->declarator_id->scope_list != NULL)
		{
		    current_class.pop();
		    $2->declarator_id->scope_list->reset_scope();
		}

		if (main_flag)
		{
		    if (main_function != NULL)
			// Log an error. Two declarations for main!!. - ecl
			;
		    main_function = function_definition_ptr;
		    main_flag = FALSE;
		}

		$$ = function_definition_ptr;
	    } 
        | TYPEDEFname                identifier.declarator 
            {
		/*
		 * A function definition is being parsed.
		 *
		 * TYPEDEFname - contains the string representation of a 
		 *		 typedef name or aggregate(class/struct/union
		 *		 /enum).
		 *
		 * identifier.declarator - the function name including the
		 * 			   parameters.
		 *
		 * compound.statement - a statement list containing the
		 *			statements in the function body.
		 */

		/*
		 * For member function definitions the corresponding class
		 * members must be in scope.  If this is a member function 
		 * definition the  identifier.declarator will have a scope_list
		 * entry. Call the member function set_scope for the scope_list
		 * entry to setup the correct scope.
		 */

		if ($2->declarator_id->scope_list != NULL)
		{
		    class_definition *class_def_ptr = 
			$2->declarator_id->scope_list->get_last();
		    if (class_def_ptr->mentat_class_flag)
		    {
			if (mentat_class == NULL)
			{
			    mentat_class = class_def_ptr;

			    // Enter the SELF record in the symbol table.
			    type *self_type_ptr = class_def_ptr->
				identifier_ptr->type_ptr->duplicate();
			    identifier *self_id_ptr = 
				new identifier("SELF",IDENTIFIER);
			    self_type_ptr->expand_typedefs();
			    self_id_ptr->type_ptr = self_type_ptr;
			    self_id_ptr = global_table.insert(self_id_ptr);
			}
			else
			    if (mentat_class != class_def_ptr)
			    {
				// Log an error,more than one mentat class with
				// member function definitions.
				cerr << "Multiple Mentat classes with member";
				cerr << " function definitons seen!\n";
			    }			    
		    }
		    
		    current_class.push(class_def_ptr);

		    $2->declarator_id->scope_list->set_scope();
		}
		
		/*
		 * The identifier declarator is a function name with
		 * parameter declarations.  Set up a symbol table containing
		 * the parameters and use this symbol table when parsing the
		 * function body.  
		 */

		$2->set_parameter_symbol_table();

		if (strcmp($2->declarator_id->id_string, "main") == 0)
		    main_flag = TRUE;
	    }
                                                            compound.statement
            {
		EXECUTE_DEBUG(cerr << "TYPEDEFname function";)
		EXECUTE_DEBUG(cerr << " definition for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Lookup the typedef name in the symbol table.  It must be in
		 * the symbol table for the lexer to have identified this as
		 * a typedef name, so no need to check the return value.
		 */

		identifier *typedef_id_ptr = scope_stack.lookup($1);
		type *typedef_type_ptr = typedef_id_ptr->type_ptr->duplicate();
		
		/*
		 * Call general_declaration to enter this function definition 
		 * in the symbol table.  Pass a duplicate of the type because 
		 * the canonical type entered with the identifier in the symbol
		 * table may have a different form than the type stored with 
		 * the  declarator object.
		 */

		general_declaration(typedef_type_ptr->duplicate(), $2,
				    NONE, TRUE);
		
		// Create a function definition object for this function.
		function_definition *function_definition_ptr = 
		    new function_definition(typedef_type_ptr, $2, 
					    (compound_statement *)$4);

		/*
		 * Set the definition field in the identifier being declared to
		 * point to the function definition object.
		 */

		$2->declarator_id->function_definition_ptr = 
		    function_definition_ptr;

		// If this is a member function reset the scope.
		if ($2->declarator_id->scope_list != NULL)
		{
		    current_class.pop();
		    $2->declarator_id->scope_list->reset_scope();
		}

		if (main_flag)
		{
		    if (main_function != NULL)
			// Log an error. Two declarations for main!!. - ecl
			;
		    main_function = function_definition_ptr;
		    main_flag = FALSE;
		}

		$$ = function_definition_ptr;
	    } 
        | declaration.qualifier.list identifier.declarator 
            {
		/*
		 * A function definition is being parsed.
		 *
		 * declaration.qualifier.list - can contain storage classes,the
		 * 				keywords typedef, volatile and 
		 *				const but no base type.
		 *
		 * identifier.declarator - the function name including the
		 * 			   parameters.
		 *
		 * compound.statement - a statement list containing the
		 *			statements in the function body.
		 */

		/*
		 * For member function definitions the corresponding class
		 * members must be in scope.  If this is a member function 
		 * definition the  identifier.declarator will have a scope_list
		 * entry. Call the member function set_scope for the scope_list
		 * entry to setup the correct scope.
		 */

		if ($2->declarator_id->scope_list != NULL)
		{
		    class_definition *class_def_ptr = 
			$2->declarator_id->scope_list->get_last();
		    if (class_def_ptr->mentat_class_flag)
		    {
			if (mentat_class == NULL)
			{
			    mentat_class = class_def_ptr;

			    // Enter the SELF record in the symbol table.
			    type *self_type_ptr = class_def_ptr->
				identifier_ptr->type_ptr->duplicate();
			    identifier *self_id_ptr = 
				new identifier("SELF",IDENTIFIER);
			    self_type_ptr->expand_typedefs();
			    self_id_ptr->type_ptr = self_type_ptr;
			    self_id_ptr = global_table.insert(self_id_ptr);
			}
			else
			    if (mentat_class != class_def_ptr)
			    {
				// Log an error,more than one mentat class with
				// member function definitions.
				cerr << "Multiple Mentat classes with member";
				cerr << " function definitons seen!\n";
			    }			    
		    }
		    
		    current_class.push(class_def_ptr);

		    $2->declarator_id->scope_list->set_scope();
		}
		
		/*
		 * The identifier declarator is a function name with
		 * parameter declarations.  Set up a symbol table containing
		 * the parameters and use this symbol table when parsing the
		 * function body.  
		 */

		$2->set_parameter_symbol_table();

		if (strcmp($2->declarator_id->id_string, "main") == 0)
		    main_flag = TRUE;
	    }
                                                             compound.statement
            {
		EXECUTE_DEBUG(cerr << "declaration qualifier list function";)
		EXECUTE_DEBUG(cerr << " definition for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Duplicate the type because the canonical type entered with 
		 * the identifier in the symbol table may have a different form 
		 * than the type stored with the declarator object.
		 */

		type *base_type_ptr = $1->duplicate();

		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

		/*
		 * Call general_declaration to enter this function definition 
		 * in the symbol table.  Do not enter overloaded operators or 
		 * conversion functions in the symbol table.
		 */

		general_declaration(base_type_ptr, $2, NONE, TRUE);
		
		// Create a function definition object for this function.
		function_definition *function_definition_ptr = 
		    new function_definition($1, $2, (compound_statement *)$4);

		/*
		 * Set the definition field in the identifier being declared to
		 * point to the function definition object.
		 */

		$2->declarator_id->function_definition_ptr = 
		    function_definition_ptr;

		// If this is a member function reset the scope.
		if ($2->declarator_id->scope_list != NULL)
		{
		    current_class.pop();
		    $2->declarator_id->scope_list->reset_scope();
		}

		if (main_flag)
		{
		    if (main_function != NULL)
			// Log an error. Two declarations for main!!. - ecl
			;
		    main_function = function_definition_ptr;
		    main_flag = FALSE;
		}

		$$ = function_definition_ptr;
	    } 
        | type.qualifier.list        identifier.declarator 
            {
		/*
		 * A function definition is being parsed.
		 *
		 * type.qualifier.list - can contain the keywords volatile and
		 * 			 const but no base type.
		 *
		 * identifier.declarator - the function name including the
		 * 			   parameters.
		 *
		 * compound.statement - a statement list containing the
		 *			statements in the function body.
		 */

		/*
		 * For member function definitions the corresponding class
		 * members must be in scope.  If this is a member function 
		 * definition the  identifier.declarator will have a scope_list
		 * entry. Call the member function set_scope for the scope_list
		 * entry to setup the correct scope.
		 */

		if ($2->declarator_id->scope_list != NULL)
		{
		    class_definition *class_def_ptr = 
			$2->declarator_id->scope_list->get_last();
		    if (class_def_ptr->mentat_class_flag)
		    {
			if (mentat_class == NULL)
			{
			    mentat_class = class_def_ptr;

			    // Enter the SELF record in the symbol table.
			    type *self_type_ptr = class_def_ptr->
				identifier_ptr->type_ptr->duplicate();
			    identifier *self_id_ptr = 
				new identifier("SELF",IDENTIFIER);
			    self_type_ptr->expand_typedefs();
			    self_id_ptr->type_ptr = self_type_ptr;
			    self_id_ptr = global_table.insert(self_id_ptr);
			}
			else
			    if (mentat_class != class_def_ptr)
			    {
				// Log an error,more than one mentat class with
				// member function definitions.
				cerr << "Multiple Mentat classes with member";
				cerr << " function definitons seen!\n";
			    }			    
		    }
		    
		    current_class.push(class_def_ptr);

		    $2->declarator_id->scope_list->set_scope();
		}
		
		/*
		 * The identifier declarator is a function name with
		 * parameter declarations.  Set up a symbol table containing
		 * the parameters and use this symbol table when parsing the
		 * function body.  
		 */

		$2->set_parameter_symbol_table();

		if (strcmp($2->declarator_id->id_string, "main") == 0)
		    main_flag = TRUE;
	    }
                                                             compound.statement
            {
		EXECUTE_DEBUG(cerr << "type qualifier list function";)
		EXECUTE_DEBUG(cerr << " definition for identifier: ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "\n";)

		/*
		 * Duplicate the type because the canonical type entered with 
		 * the identifier in the symbol table may have a different form 
		 * than the type stored with the declarator object.
		 */

		type *base_type_ptr = $1->duplicate();

		// Add the default base type int.
		base_type_ptr->add_type_element(new base_type(INT));

		/*
		 * Call general_declaration to enter this function definition 
		 * in the symbol table.  Do not enter overloaded operators or 
		 * conversion functions in the symbol table.
		 */

		general_declaration(base_type_ptr, $2, NONE, TRUE);
		
		// Create a function definition object for this function.
		function_definition *function_definition_ptr = 
		    new function_definition($1, $2, (compound_statement *)$4);

		/*
		 * Set the definition field in the identifier being declared to
		 * point to the function definition object.
		 */

		$2->declarator_id->function_definition_ptr = 
		    function_definition_ptr;

		// If this is a member function reset the scope.
		if ($2->declarator_id->scope_list != NULL)
		{
		    current_class.pop();
		    $2->declarator_id->scope_list->reset_scope();
		}

		if (main_flag)
		{
		    if (main_function != NULL)
			// Log an error. Two declarations for main!!. - ecl
			;
		    main_function = function_definition_ptr;
		    main_flag = FALSE;
		}

		$$ = function_definition_ptr;
	    } 
        ;

/******************************************************************************
 *
 *				 Destructors
 *
 *****************************************************************************/
    
/* 
 * I needed to add some storage class options to the constructors 
 * and  destructors  (e.g.: virtual, inline, ...  ) These were added 
 * using declaration.qualifier.list, but they should be semantically 
 * verified to be inline or virtual ONLY. If I don't have  to  allow 
 * BOTH,  then  I  COULD be explicit (and not increase ambiguities). 
 * With this LR(1) grammar, I have to reduce the  list  to  standard 
 * declaration.qualifier.list ASAP.
 */

destructor.function:
        declaration.qualifier.list '~' TYPEDEFname void.opt.param.list
                        destructor.function.body
            {
		identifier *id_ptr = new identifier($3, FUNCTION);
		id_ptr->destructor_flag = TRUE;

		declarator *decl_ptr = new declarator(id_ptr);
		decl_ptr->declarator_type = 
		    new type(new type_modifier(FUNCTION, $4));
		$$ = new function_definition($1, decl_ptr,
					     (compound_statement *)$5);
	    }
        | '~' TYPEDEFname void.opt.param.list destructor.function.body
            {
		identifier *id_ptr = new identifier($2, FUNCTION);
		id_ptr->destructor_flag = TRUE;

		declarator *decl_ptr = new declarator(id_ptr);
		decl_ptr->declarator_type = 
		    new type(new type_modifier(FUNCTION, $3));
		$$=new function_definition(decl_ptr, (compound_statement *)$4);
	    }
        ;

void.opt.param.list:
        '('        ')' type.qualifier.list.opt
	    {
		$$ = new parameter_list(new parameter());
	    }
        | '(' VOID ')' type.qualifier.list.opt
	    {
		$$ = new parameter_list(new parameter());
	    }
        ;

destructor.function.body:
        /* an actual destructor definition */
        compound.statement member.function.definition.semi.opt
        | ';'               /* only a destructor declaration */
            {
		$$ = NULL;
	    }
        ;

/******************************************************************************
 *
 *				 Constructors
 *
 *****************************************************************************/

/* Semantically  verify that the following identifier.declarator 
 * are:
 *  
 *      TYPEDEFname :: TYPEDEFname (parameter.type.list).  
 *
 * We use the more general form to prevent a clash  with  a  typical 
 * function  definition  (which  won't have a constructor.init.list) 
 * The ONLY valid declaration qualifier is INLINE.
 */

constructor.function.definition:
                                     identifier.declarator  
            {
		/*
		 * A constructor function definition is being parsed.
		 *
		 * identifier.declarator - the constructor name including the
		 * 			   parameters and csu name modifier.
		 *
		 * constructor.init.list - a ctor initializer list containing
		 *			   the initializations for non-static
		 *			   members and direct base classes.
		 *
		 * compound.statement - a statement list containing the
		 *			statements in the function body.
		 */

		/*
		 * The class members must be in scope.  Since this is a member 
		 * function definition the  identifier.declarator will have a 
		 * scope_list entry. Call the member function set_scope for the
		 * scope_list entry to setup the correct scope.
		 */

		$1->declarator_id->scope_list->set_scope();
		
		/*
		 * The identifier declarator is a function name with
		 * parameter declarations.  Set up a symbol table containing
		 * the parameters and use this symbol table when parsing the
		 * function body.  
		 */

		$1->set_parameter_symbol_table();
	    }
                                constructor.init.list compound.statement
            {
		EXECUTE_DEBUG(cerr << "Constructor definition for class ";)
		EXECUTE_DEBUG(cerr << $1->declarator_id->id_string << "!\n";)

		/*
		 * This is a constructor so its return type is a pointer to its 
		 * csu.  Lookup the csu name in the symbol table and use the
		 * type pointer to the csu type as the type.
		 */

		identifier *csu_identifier_ptr = 
		    scope_stack.lookup($1->declarator_id->id_string);
		if (csu_identifier_ptr == NULL)
		{
		    // Log an error. -ecl
		}
		type *type_ptr = csu_identifier_ptr->type_ptr;
		
		// Create a function definition object for this constructor.
		function_definition *ctor_definition_ptr = 
		    new function_definition(NULL, $1,(compound_statement *)$4);

		// Set the field for the constructor initializers.
		ctor_definition_ptr->ctor_initializers = $3;

		// This is a member function so reset the scope.
		$1->declarator_id->scope_list->reset_scope();
		
		$$ = ctor_definition_ptr;
	    }
        | declaration.qualifier.list identifier.declarator  
            {
		/*
		 * The identifier declarator is a function name with
		 * parameter declarations.  Set up a symbol table containing
		 * the parameters and use this for the symbol table for the
		 * function body.  This is a member function declaration thus 
		 * the identifier declarator may have a scope_list entry.  Use
		 * this list to set up the symbol table stack correctly.
		 */

		$2->set_parameter_symbol_table();

		if ($2->declarator_id->scope_list != NULL)
		    $2->declarator_id->scope_list->set_scope();

		// Set the inline flag in the identifier.
		$2->declarator_id->inline_flag = TRUE;
		
	    }
                                constructor.init.list compound.statement
            {
		EXECUTE_DEBUG(cerr << "Constructor definition for ";)
		EXECUTE_DEBUG(cerr << $2->declarator_id->id_string << "!\n";)

		/*
		 * This is a constructor so it returns a pointer to its object.
		 * Come back to this - ecl. Set the type to NULL for now.
		 * Need to lookup the class name in the symbol table and get 
		 * its type.
		 */

		type *base_type_ptr = NULL;
		
		function_definition *ctor_func_def = 
		    new function_definition($1, $2, (compound_statement *)$5);

		ctor_func_def->ctor_initializers = $4;
		
		// This is a member function so pop the extra symbol tables.
		if ($2->declarator_id->scope_list != NULL)
		    $2->declarator_id->scope_list->reset_scope();
		
		$$ = ctor_func_def;
	    }
        ;

/*
 * The following use of declaration.specifiers are made to  allow 
 * for  a TYPEDEFname preceded by an INLINE modifier. This fact must 
 * be verified semantically.  It should also be  verified  that  the 
 * TYPEDEFname  is  ACTUALLY  the  class name being elaborated. Note 
 * that we could break out typedef.declaration.specifier from within 
 * declaration.specifier, and we  might  narrow  down  the  conflict 
 * region a bit. A second alternative (to what is done) for cleaning 
 * up  this  stuff  is  to  let the tokenizer specially identify the 
 * current class being elaborated as a special token, and not just a 
 * typedef.name. Unfortunately, things would get very confusing  for 
 * the  lexer,  as  we may pop into enclosed tag elaboration scopes; 
 * into function definitions; or into both recursively! 
 */
    
constructor.function.in.class:
        declaration.specifier   constructor.parameter.list.and.body
            {
		/*
		 * This declaration.specifier must contain const or volatile
		 * plus the TYPEDEFname.  Deal with this case later - ecl.
		 */
	    }
        | TYPEDEFname           constructor.parameter.list.and.body
            {
		/*
		 * Build the string _TYPEDEFname_ctor and use this as the
		 * function name for this constructor function.
		 */
		$2->function_name->declarator_id = new identifier($1,FUNCTION);

		$$ = $2;
	    }
        ;

/*
 * The following conflicts with member declarations-
 * constructor.parameter.list.and.body:
 *       parameter.type.list ';'
 *       | parameter.type.list constructor.init.list.opt compound.statement
 *       ;
 * so parameter.type.list was expanded inline to get 
 */

constructor.parameter.list.and.body:
          '('                           ')' type.qualifier.list.opt ';'
            {
		// Build the parameter type list.  In this case it is empty.
		parameter_list *param_list_ptr = new parameter_list(new 
								 parameter());

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' type.name initializer     ')' type.qualifier.list.opt ';'
            {
		// Build the parameter type list.  In this case it is empty.
		parameter_list *param_list_ptr = new parameter_list(new 
							    parameter($2, $3));

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' named.parameter.type.list ')' type.qualifier.list.opt ';'
            {
		// Build the parameter type list.  In this case it is empty.
		parameter_list *param_list_ptr = $2;

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '('                           ')' type.qualifier.list.opt
                constructor.init.list.opt compound.statement 
		member.function.definition.semi.opt
		
            {
		// Build the parameter type list.  In this case it is empty.
		parameter_list *param_list_ptr = new parameter_list(new 
								 parameter());

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr), 
					     (compound_statement *)$5);
		$$->ctor_initializers = $4;
	    }
        | '(' type.name initializer     ')' type.qualifier.list.opt 
            {
		// Build the parameter type list.  In this case it is empty.
		parameter_list *param_list_ptr = new parameter_list(new 
							    parameter($2, $3));

		/*
		 * Set up a symbol table for the parameters and use this when
		 * parsing the constructor body.
		 */

		param_list_ptr->set_parameter_symbol_table();

		$<parameter_list_ptr>$ = param_list_ptr;
	    }
                constructor.init.list.opt compound.statement  
		member.function.definition.semi.opt	
            {
		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
						 $<parameter_list_ptr>6));
		
		$$ = new function_definition(new declarator(type_ptr), 
					     (compound_statement *)$8);
		$$->ctor_initializers = $7;
	    }
	| '(' named.parameter.type.list ')' type.qualifier.list.opt 
            {
		// Build the parameter type list.  In this case it is empty.
		parameter_list *param_list_ptr = $2;

		/*
		 * Set up a symbol table for the parameters and use this when
		 * parsing the constructor body.
		 */

		param_list_ptr->set_parameter_symbol_table();		

		$<parameter_list_ptr>$ = param_list_ptr;

	    }
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
						 $<parameter_list_ptr>5));
		
		$$ = new function_definition(new declarator(type_ptr), 
					     (compound_statement *)$7);
		$$->ctor_initializers = $6;
	    }
        | constructor.conflicting.parameter.list.and.body
        ;

/*
 * The following conflicted with member declaration-
 * constructor.conflicting.parameter.list.and.body:
 *     '(' type.name ')'                 type.qualifier.list.opt ';'
 *     | '(' type.name ')'                 type.qualifier.list.opt
 *             constructor.init.list.opt compound.statement
 *     ;
 * so type.name was inline expanded to get the following... 
 */

/*
 * The type.qualifier.list.opt is for const and volatile constructors.
 * Ignore it for now - ecl.
 */

constructor.conflicting.parameter.list.and.body:
        '(' type.specifier ')' type.qualifier.list.opt
                ';'
            {
		// Build the parameter.
		parameter *param_ptr = new parameter(
						new variable_declaration($2));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' basic.type.name  ')' type.qualifier.list.opt
                ';'
            {
		// Build the parameter.
		parameter *param_ptr = new parameter(
				      new variable_declaration(new type($2)));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' TYPEDEFname  ')' 
                ';'
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();

		// Build the parameter.
		parameter *param_ptr = new parameter(
					new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' TYPEDEFname  ')' type.qualifier.list
                ';'
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();

		// Build the parameter.
		parameter *param_ptr = new parameter(
					new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' type.qualifier.list  ')' type.qualifier.list.opt
                ';'
            {
		// Build the parameter.
		parameter *param_ptr = new parameter(
						new variable_declaration($2));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' type.specifier abstract.declarator ')' type.qualifier.list.opt
                ';'
            {
		// Build the parameter.
		parameter *param_ptr = new parameter(
			     new variable_declaration($2, new declarator($3)));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' basic.type.name abstract.declarator ')' type.qualifier.list.opt
                ';'
            {
		// Build the parameter.
		parameter *param_ptr = new parameter(
		   new variable_declaration(new type($2), new declarator($3)));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' type.qualifier.list abstract.declarator ')' type.qualifier.list.opt
                ';'
            {
		// Build the parameter.
		parameter *param_ptr = new parameter(
			     new variable_declaration($2, new declarator($3)));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }

        | '(' type.specifier ')' type.qualifier.list.opt
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the parameter.
		parameter *param_ptr = new parameter(
						new variable_declaration($2));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$6);
		$$->ctor_initializers = $5;
	    }
        | '(' basic.type.name  ')' type.qualifier.list.opt
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the parameter.
		parameter *param_ptr = new parameter(
				      new variable_declaration(new type($2)));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$6);
		$$->ctor_initializers = $5;
	    }
        | '(' TYPEDEFname  ')' 
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();

		// Build the parameter.
		parameter *param_ptr = new parameter(
					new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$5);
		$$->ctor_initializers = $4;
	    }
        | '(' TYPEDEFname  ')' type.qualifier.list
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();

		// Build the parameter.
		parameter *param_ptr = new parameter(
					new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$6);
		$$->ctor_initializers = $5;
	    }
        | '(' type.qualifier.list  ')' type.qualifier.list.opt
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the parameter.
		parameter *param_ptr = new parameter(
						new variable_declaration($2));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$6);
		$$->ctor_initializers = $5;
	    }
        | '(' type.specifier abstract.declarator ')' type.qualifier.list.opt 
	    {
		// Build the parameter.
		parameter *param_ptr = new parameter(
			     new variable_declaration($2, new declarator($3)));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		/*
		 * Set up a symbol table for the parameters and use this when
		 * parsing the constructor body.
		 */

		param_list_ptr->set_parameter_symbol_table();
		
		$<parameter_list_ptr>$ = param_list_ptr;
	    }
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
						      $<parameter_list_ptr>5));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$8);
		$$->ctor_initializers = $7;
	    }
        | '(' basic.type.name abstract.declarator ')' type.qualifier.list.opt
	    {
		// Build the parameter.
		parameter *param_ptr = new parameter(
		   new variable_declaration(new type($2), new declarator($3)));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		/*
		 * Set up a symbol table for the parameters and use this when
		 * parsing the constructor body.
		 */

		param_list_ptr->set_parameter_symbol_table();
		
		$<parameter_list_ptr>$ = param_list_ptr;
	    }
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
						  $<parameter_list_ptr>6));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$8);
		$$->ctor_initializers = $7;
	    }
        | '(' type.qualifier.list abstract.declarator ')' 
	                                               type.qualifier.list.opt 
	    {
		// Build the parameter.
		parameter *param_ptr = new parameter(
			     new variable_declaration($2, new declarator($3)));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		/*
		 * Set up a symbol table for the parameters and use this when
		 * parsing the constructor body.
		 */

		param_list_ptr->set_parameter_symbol_table();

		$<parameter_list_ptr>$ = param_list_ptr;
	    }
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
						  $<parameter_list_ptr>6));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$8);
		$$->ctor_initializers = $7;
	    }
        | constructor.conflicting.typedef.declarator
        ;


/*
 * The following have ambiguities with member declarations-
 * constructor.conflicting.typedef.declarator:
 *   '(' TYPEDEFname abstract.declarator ')' type.qualifier.list.opt
 *             ';'
 *   |  '(' TYPEDEFname abstract.declarator ')' type.qualifier.list.opt
 *             constructor.init.list.opt compound.statement
 *   ;
 * which can be deferred by expanding abstract.declarator, and in two 
 * cases parameter.qualifier.list, resulting in ...
 */

constructor.conflicting.typedef.declarator:
        '(' TYPEDEFname unary.abstract.declarator ')' type.qualifier.list.opt
                ';'
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		typedef_ptr->add_type($3);

		// Build the parameter.
		parameter *param_ptr = new parameter(
				new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' TYPEDEFname unary.abstract.declarator ')' type.qualifier.list.opt
	    {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		$3->add_type(typedef_ptr);

		// Build the parameter.
		parameter *param_ptr = new parameter(
				new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		/*
		 * Set up a symbol table for the parameters and use this when
		 * parsing the constructor body.
		 */

		param_list_ptr->set_parameter_symbol_table();

		$<parameter_list_ptr>$ = param_list_ptr;
	    }
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
						$<parameter_list_ptr>6));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$8);
		$$->ctor_initializers = $7;
	    }

        | '(' TYPEDEFname postfix.abstract.declarator ')' 
	                                                type.qualifier.list.opt
                ';'
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		$3->add_type(typedef_ptr);

		// Build the parameter.
		parameter *param_ptr = new parameter(
				new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' TYPEDEFname postfix.abstract.declarator ')' 
	                                              type.qualifier.list.opt 
	    {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		$3->add_type(typedef_ptr);

		// Build the parameter.
		parameter *param_ptr = new parameter(
				new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		/*
		 * Set up a symbol table for the parameters and use this when
		 * parsing the constructor body.
		 */

		param_list_ptr->set_parameter_symbol_table();

		$<parameter_list_ptr>$ = param_list_ptr;
	    }
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
						     $<parameter_list_ptr>6));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$8);
		$$->ctor_initializers = $7;
	    }

        | '(' TYPEDEFname postfixing.abstract.declarator ')' 
	                                                    type.qualifier.list
                ';'
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		$3->add_type(typedef_ptr);

		// Build the parameter.
		parameter *param_ptr = new parameter(
				new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' TYPEDEFname postfixing.abstract.declarator ')' 
	                                                   type.qualifier.list 
	    {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		$3->add_type(typedef_ptr);

		// Build the parameter.
		parameter *param_ptr = 
		    new parameter(new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		/*
		 * Set up a symbol table for the parameters and use this when
		 * parsing the constructor body.
		 */

		param_list_ptr->set_parameter_symbol_table();

		$<parameter_list_ptr>$ = param_list_ptr;
	    }
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
						$<parameter_list_ptr>6));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$8);
		$$->ctor_initializers = $7;
	    }

        | '(' TYPEDEFname postfixing.abstract.declarator ')' 
                ';'
            {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		$3->add_type(typedef_ptr);

		// Build the parameter.
		parameter *param_ptr = new parameter(
				new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
							    param_list_ptr));
		
		$$ = new function_definition(new declarator(type_ptr));
	    }
        | '(' TYPEDEFname postfixing.abstract.declarator ')'  
	    {
		// Lookup the typedef name in the symbol table.
		identifier *id_ptr = scope_stack.lookup($2);
		type *typedef_ptr = id_ptr->type_ptr->duplicate();
		$3->add_type(typedef_ptr);

		// Build the parameter.
		parameter *param_ptr = new parameter(
				new variable_declaration(typedef_ptr));

		// Build the parameter list.
		parameter_list *param_list_ptr = new parameter_list(param_ptr);

		/*
		 * Set up a symbol table for the parameters and use this when
		 * parsing the constructor body.
		 */

		param_list_ptr->set_parameter_symbol_table();

		$<parameter_list_ptr>$ = param_list_ptr;
	    }
                constructor.init.list.opt compound.statement
		member.function.definition.semi.opt
            {
		// Build the type, which for a constructor is FUNCTION.
		type *type_ptr = new type(new type_modifier(FUNCTION, 
						     $<parameter_list_ptr>5));
		
		$$ = new function_definition(new declarator(type_ptr),
					     (compound_statement *)$7);
		$$->ctor_initializers = $6;
	    }
        ;


constructor.init.list.opt:
        /* nothing */
	    {
		$$ = NULL;
	    }
        | constructor.init.list
	    {
		$$ = $1;
	    }
        ;

constructor.init.list:
        ':' constructor.init
	    {
		$$ = new ctor_initializer_list($2);
	    }
        | constructor.init.list ',' constructor.init
	    {
		$$ = $1;
		$$->add_ctor_initializer($3);
	    }
        ;

constructor.init:
	identifier.or.typedef.name   '(' argument.expression.list ')'
	    {
		$$ = new ctor_initializer($1, $3);
	    }
        | identifier.or.typedef.name '('                          ')'
	    {
		$$ = new ctor_initializer($1, NULL);
	    }
        | '(' argument.expression.list ')' /* Single inheritance ONLY*/
	    {
		$$ = new ctor_initializer(NULL, $2);
	    }
        | '(' ')' /* Is this legal? It might be default! */
	    {
		$$ = new ctor_initializer(NULL, NULL);
	    }
        ;        

/******************************************************************************
 *
 *		      		Declarators
 *
 *****************************************************************************/

declarator:
        typedef.declarator
        | identifier.declarator
        ;
        
/******************************************************************************
 *
 *				 Typedef Declarators
 *
 *****************************************************************************/

typedef.declarator:
        paren.typedef.declarator  	  /* would be ambiguous as parameter*/
        | simple.paren.typedef.declarator /* also ambiguous                 */
        | parameter.typedef.declarator    /* not ambiguous as param         */
        ;

parameter.typedef.declarator:
        TYPEDEFname
            {
		/*
		 * A typedef name is being redefined in a different scope.
		 * Treat it as a new identifier.
		 */

		$$ = new declarator(new identifier($1, IDENTIFIER));
	    }
        | TYPEDEFname postfixing.abstract.declarator
            {
		/*
		 * A typedef name is being redefined in a different scope.
		 * Treat it as a new identifier.
		 */

		$$ = new declarator(new identifier($1, IDENTIFIER));
		$$->declarator_type = $2;
	    }
        | clean.typedef.declarator
        ;

/*
 * The following have at  least  one  '*' or  '&'.   There  is  no 
 * (redundant) '(' between the '*'/'&' and the TYPEDEFname. 
 */
    
clean.typedef.declarator:
        clean.postfix.typedef.declarator
        | indirect.or.reference parameter.typedef.declarator
            {
        	$2->add_type(new type(new type_modifier($1)));
		$$ = $2;
	    }
        | pointer.operator      parameter.typedef.declarator  
            {
		$2->add_type($1);
		$$ = $2;
	    }
        ;

clean.postfix.typedef.declarator:
        '(' clean.typedef.declarator ')'
            {
		$$ = $2;
	    }
        | '(' clean.typedef.declarator ')' postfixing.abstract.declarator
            {
		$2->declarator_type->add_type($4);
		$$ = $2;
	    }
        ;

/*
 * The  following have a redundant '(' placed immediately to the 
 * left of the TYPEDEFname 
 */
    
paren.typedef.declarator:
        paren.postfix.typedef.declarator
        | indirect.or.reference '(' simple.paren.typedef.declarator ')'
            {
		$3->add_type(new type(new type_modifier($1)));
		$$ = $3;
	    }
        | pointer.operator      '(' simple.paren.typedef.declarator ')'
            {
		$3->add_type($1);
		$$ = $3;
	    }
        | indirect.or.reference '(' TYPEDEFname ')' /* redundant paren */
            {
		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($3, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($3);
		
		$$ = new declarator(id_ptr, new type(new type_modifier($1)));
	    }
        | pointer.operator      '(' TYPEDEFname ')' /* redundant paren */
            {
		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($3, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($3);
		
		$$ = new declarator(id_ptr, $1);
	    }
        | indirect.or.reference paren.typedef.declarator
            {
		$2->add_type(new type(new type_modifier($1)));
		$$ = $2;
	    }
        | pointer.operator      paren.typedef.declarator
            {
		$2->add_type($1);
		$$ = $2;
	    }
        ;
        
paren.postfix.typedef.declarator:
        '(' paren.typedef.declarator ')'
            {
		$$ = $2;
	    }
        | '(' simple.paren.typedef.declarator postfixing.abstract.declarator ')' 
            {	
		$3->add_type($2->declarator_type);
		$2->declarator_type = $3;
		$$ = $2;
	    }
        | '(' TYPEDEFname postfixing.abstract.declarator ')' /*redundant par */
            {
		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($2, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($2);
		
		$$ = new declarator(id_ptr, $3);
	    }
        | '(' paren.typedef.declarator ')' postfixing.abstract.declarator
            {
		$2->declarator_type->add_type($4);
		$$ = $2;
	    }
        ;

/*
 * The following excludes lone TYPEDEFname to help in a  conflict 
 * resolution.   We  have  special cased lone TYPEDEFname along side 
 * all uses 
 */
    
simple.paren.typedef.declarator:
        '(' TYPEDEFname ')'
            {
		/*
		 * A typedef name is being redefined in a different scope.
		 * Enter the name in the current symbol table.
		 */

		identifier *id_ptr = scope_stack.insert($2, IDENTIFIER);
		if (id_ptr == NULL)
		    // Log an error, the identifier is being redeclared.
		    id_redeclared_error($2);
		
		$$ = new declarator(id_ptr);
	    }
        | '(' simple.paren.typedef.declarator ')'
            {
		$$ = $2;
	    }
        ;

/******************************************************************************
 *
 *				 Identifier Declarators
 *
 *****************************************************************************/

identifier.declarator:
        unary.identifier.declarator  
        | paren.identifier.declarator
        ;

/*
 * The following allows "function return array  of"  as  well  as 
 * "array  of  function  returning".  It COULD be cleaned up the way 
 * abstract declarators have been.  This change might make  it  hard 
 * to  recover from user's syntax errors, whereas now they appear as 
 * simple semantic errors. 
 */

unary.identifier.declarator:
        postfix.identifier.declarator
        | indirect.or.reference identifier.declarator
            {
		$2->add_type(new type(new type_modifier($1)));
		$$ = $2;
	    }
        | pointer.operator      identifier.declarator
            {
		$2->add_type($1);
		$$ = $2;
	    }
        ;

postfix.identifier.declarator:
        paren.identifier.declarator postfixing.abstract.declarator
            {
		$1->declarator_type = $2;
		$$ = $1;
	    }
        | '(' unary.identifier.declarator ')'
            {
		$$ = $2;
	    }
        | '(' unary.identifier.declarator ')' postfixing.abstract.declarator
            {
		$2->add_type($4);
		$$ = $2;
	    }
        ;

/******************************************************************************
 *
 *				 Abstract Declarators
 *
 *****************************************************************************/

abstract.declarator:
        unary.abstract.declarator
        | postfix.abstract.declarator
        | postfixing.abstract.declarator
        ;

postfixing.abstract.declarator:
        array.abstract.declarator
        | parameter.type.list
            {
		$$ = new type(new type_modifier(FUNCTION, $1));
	    }
        ;

array.abstract.declarator:
        '[' ']'
            {
		$$ = new type(new type_modifier(ARRAY));
	    }
        | '[' constant.expression ']'
            {
		$$ = new type(new type_modifier(ARRAY, $2));
	    }
        | array.abstract.declarator '[' constant.expression ']'
            {
		$1->add_type_element(new type_modifier(ARRAY, $3));
		$$ = $1;
	    }
        ;

unary.abstract.declarator:
        indirect.or.reference 
            {
		$$ = new type(new type_modifier($1));
	    }
        | pointer.operator 
        | indirect.or.reference abstract.declarator
            {
		$2->add_type_element(new type_modifier($1));
		$$ = $2;
	    }
        | pointer.operator      abstract.declarator
            {
		$2->add_type($1);
		$$ = $2;
	    }
        ;

postfix.abstract.declarator:
        '(' unary.abstract.declarator ')'
            {    
		$$ = $2;
	    }
        | '(' postfix.abstract.declarator ')'
            {    
		$$ = $2;
	    }
        | '(' postfixing.abstract.declarator ')'
            {    
		$$ = $2;
	    }
        | '(' unary.abstract.declarator ')' postfixing.abstract.declarator
            {    
		$2->add_type($4);
		$$ = $2;
	    }
        ;

indirect.or.reference:
        '*'
            {
		$$ = $<token>1;
	    }
        | '&'
            {
		$$ = $<token>1;
	    }
        ;

pointer.operator:
        TYPEDEFname CLCL '*' type.qualifier.list.opt
            {

		/*
		 * Don't know what to do here yet!! Abort for now. - ecl
		 */

		cerr << "Rule pointer.operator: #1 not yet implemented!!\n";
		exit(-1);
	    }
        | indirect.or.reference type.qualifier.list
            {
		$$ = $2;
		$$->add_type_element(new type_modifier($1));
	    }
        ;
%%

char *op_to_string(TOKEN op)
{
    switch (op)
    {
      case '+':
	return "+";
      case '-':
	return "-";
      case '*':
	return "*";
      case '/':
	return "/";
      case '%':
	return "%";
      case '^':
	return "^";
      case '&':
	return "&";
      case '|':
	return "|";
      case '~':
	return "~";
      case '!':
	return "!";
      case '<':
	return "<";
      case '>':
	return ">";
      case LS:
	return "<<";
      case RS:
	return ">>";
      case ANDAND:
	return "&&";
      case OROR:
	return "||";
      case ARROW:
	return "->";
      case ICR:
	return "++";
      case DECR:
	return "--";
      case LE:
	return "<=";
      case GE:
	return ">=";
      case EQ:
	return "==";
      case NE:
	return "!=";
      case '(':
	return "()";
      case '[':
	return "[]";
      case NEW:
	return "new";
      case DELETE:
	return "delete";
      case ',':
	return ",";
      case '=':
	return "=";
      case MULTassign:
	return "*=";
      case DIVassign:
	return "/=";
      case PLUSassign:
	return "+=";
      case MINUSassign:
	return "-=";
      case LSassign:
	return "<<=";
      case RSassign:
	return ">>=";
      case ANDassign:
	return "&=";
      case ERassign:
	return "^=";
      case ORassign:
	return "|=";
      default:
	//Log an error.
	;
    }
}

boolean match_token(TOKEN token1, TOKEN token2)
{
    if (token1 == UNION && token2 == UNION)
	return TRUE;
    if (token1 == STRUCT && (token2 == STRUCT || token2 == CLASS))
	return TRUE;
    if (token1 == CLASS && (token2 == STRUCT || token2 == CLASS))
	return TRUE;
    return FALSE;
}
